# flake8: noqa
"""This is a standalone scripts shows which models have permissions."""

import inspect
import os
import sys

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(current_dir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)
import django

# init and setup django project
print("Initializing Django...")
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

from rich.pretty import pprint

from django.apps import apps

from memberaudit.models import General
from memberaudit.tests.utils import permissions_for_model


def main():
    my_app_label = General._meta.app_label
    my_models = [o for o in apps.get_models() if o._meta.app_label == my_app_label]
    my_models.sort(key=lambda o: o._meta.model_name)
    results = {}
    for model_class in my_models:
        has_permissions = permissions_for_model(model_class).exists()
        results[model_class.__name__] = has_permissions

    pprint(results)


if __name__ == "__main__":
    main()



import datetime as dt

from django.test import TestCase
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from app_utils.testing import NoSocketsTestCase

from memberaudit.constants import EveFactionId
from memberaudit.models import (
    CharacterContract,
    CharacterFwStats,
    CharacterWalletJournalEntry,
    Location,
)
from memberaudit.tests.testdata.factories import (
    create_character_contract,
    create_character_contract_item,
    create_character_fw_stats,
    create_character_ship,
    create_character_standing,
    create_character_title,
    create_character_wallet_journal_entry,
    create_eve_market_price,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import create_memberaudit_character


class TestCharacterShip(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.user = cls.character_1001.eve_character.character_ownership.user

    def test_str(self):
        # given
        create_character_ship(
            character=self.character_1001, eve_type=EveType.objects.get(id=603)
        )
        # when
        result = str(self.character_1001.ship)
        # then
        self.assertIn("Bruce Wayne", result)
        self.assertIn("Merlin", result)


class TestCharacterContract(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.character_1002 = create_memberaudit_character(1002)
        cls.token = (
            cls.character_1001.eve_character.character_ownership.user.token_set.first()
        )
        cls.jita = EveSolarSystem.objects.get(id=30000142)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.amamake = EveSolarSystem.objects.get(id=30002537)
        cls.structure_1 = Location.objects.get(id=1000000000001)
        cls.snake_alpha_type = EveType.objects.get(name="High-grade Snake Alpha")
        cls.snake_beta_type = EveType.objects.get(name="High-grade Snake Beta")

    def setUp(self) -> None:
        self.contract = create_character_contract(
            character=self.character_1001,
            contract_id=42,
            availability=CharacterContract.AVAILABILITY_PERSONAL,
            contract_type=CharacterContract.TYPE_ITEM_EXCHANGE,
            date_issued=now(),
            date_expired=now() + dt.timedelta(days=3),
            for_corporation=False,
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_OUTSTANDING,
            start_location=self.jita_44,
            end_location=self.jita_44,
        )
        self.contract_completed = create_character_contract(
            character=self.character_1001,
            contract_id=43,
            availability=CharacterContract.AVAILABILITY_PERSONAL,
            contract_type=CharacterContract.TYPE_ITEM_EXCHANGE,
            date_issued=now() - dt.timedelta(days=3),
            date_completed=now() - dt.timedelta(days=2),
            date_expired=now() - dt.timedelta(days=1),
            for_corporation=False,
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_FINISHED,
            start_location=self.jita_44,
            end_location=self.jita_44,
        )

    def test_str(self):
        self.assertEqual(str(self.contract), f"{self.character_1001}-42")

    def test_is_completed(self):
        self.assertFalse(self.contract.is_completed)
        self.assertTrue(self.contract_completed.is_completed)

    def test_has_expired(self):
        self.assertFalse(self.contract.has_expired)
        self.assertTrue(self.contract_completed.has_expired)

    def test_hours_issued_2_completed(self):
        self.assertIsNone(self.contract.hours_issued_2_completed)
        self.assertEqual(self.contract_completed.hours_issued_2_completed, 24)

    def test_summary_one_item_1(self):
        create_character_contract_item(
            contract=self.contract,
            record_id=1,
            is_included=True,
            is_singleton=False,
            quantity=1,
            eve_type=self.snake_alpha_type,
        )
        self.assertEqual(self.contract.summary(), "High-grade Snake Alpha")

    def test_summary_one_item_2(self):
        create_character_contract_item(
            contract=self.contract,
            record_id=1,
            is_included=True,
            is_singleton=False,
            quantity=1,
            eve_type=self.snake_alpha_type,
        )
        create_character_contract_item(
            contract=self.contract,
            record_id=2,
            is_included=False,
            is_singleton=False,
            quantity=1,
            eve_type=self.snake_beta_type,
        )
        self.assertEqual(self.contract.summary(), "High-grade Snake Alpha")

    def test_summary_multiple_item(self):
        create_character_contract_item(
            contract=self.contract,
            record_id=1,
            is_included=True,
            is_singleton=False,
            quantity=1,
            eve_type=self.snake_alpha_type,
        ),
        create_character_contract_item(
            contract=self.contract,
            record_id=2,
            is_included=True,
            is_singleton=False,
            quantity=1,
            eve_type=self.snake_beta_type,
        )
        self.assertEqual(self.contract.summary(), "[Multiple Items]")

    def test_summary_no_items(self):
        self.assertEqual(self.contract.summary(), "(no items)")

    def test_can_calculate_pricing_1(self):
        """calculate price and total for normal item"""
        create_character_contract_item(
            contract=self.contract,
            record_id=1,
            is_included=True,
            is_singleton=False,
            quantity=2,
            eve_type=self.snake_alpha_type,
        ),
        create_eve_market_price(eve_type=self.snake_alpha_type, average_price=5000000)
        qs = self.contract.items.annotate_pricing()
        item_1 = qs.get(record_id=1)
        self.assertEqual(item_1.price, 5000000)
        self.assertEqual(item_1.total, 10000000)

    def test_can_calculate_pricing_2(self):
        """calculate price and total for BPO"""
        create_character_contract_item(
            contract=self.contract,
            record_id=1,
            is_included=True,
            is_singleton=False,
            quantity=1,
            raw_quantity=-2,
            eve_type=self.snake_alpha_type,
        ),
        create_eve_market_price(eve_type=self.snake_alpha_type, average_price=5000000)
        qs = self.contract.items.annotate_pricing()
        item_1 = qs.get(record_id=1)
        self.assertIsNone(item_1.price)
        self.assertIsNone(item_1.total)


class TestCharacterFwStatsRankNameGeneric(TestCase):
    def test_should_return_rank_name_when_found(self):
        # when
        result = CharacterFwStats.rank_name_generic(EveFactionId.CALDARI_STATE, 4)
        # then
        self.assertEqual(result, "Major")

    def test_should_raise_error_for_unknown_faction(self):
        # when/then
        with self.assertRaises(ValueError):
            CharacterFwStats.rank_name_generic(42, 4)

    def test_should_raise_error_for_invalid_rank(self):
        # when/then
        with self.assertRaises(ValueError):
            CharacterFwStats.rank_name_generic(EveFactionId.CALDARI_STATE, 42)


class TestCharacterFwStatsRankNameObject(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1121)

    def test_should_return_rank_name_when_found(self):
        # given
        obj = create_character_fw_stats(character=self.character, current_rank=4)
        # when/then
        self.assertEqual(obj.current_rank_name(), "Major")

    def test_should_return_rank_name_when_not_found(self):
        # given
        obj = create_character_fw_stats(character=self.character, faction=None)
        # when/then
        self.assertEqual(obj.current_rank_name(), "")


class TestCharacterStanding(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_effective_standing_with_connections(self):
        # given
        eve_entity = EveEntity.objects.get(id=1901)
        obj = create_character_standing(self.character, eve_entity, standing=4.99)
        # when
        result = obj.effective_standing(3, 0, 0)
        # then
        self.assertAlmostEqual(result, 5.59, 2)

    def test_effective_standing_with_diplomacy(self):
        # given
        eve_entity = EveEntity.objects.get(id=1901)
        obj = create_character_standing(self.character, eve_entity, standing=-4.76)
        # when
        result = obj.effective_standing(0, 0, 5)
        # then
        self.assertAlmostEqual(result, -1.81, 2)


class TestCharacterTitle(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_return_str(self):
        # given
        obj = create_character_title(character=self.character, name="Dummy")
        # when
        result = str(obj)
        # then
        self.assertIn("Dummy", result)


class TestCharacterWalletJournals(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_return_eve_entity_ids(self):
        # given
        obj = create_character_wallet_journal_entry(
            character=self.character, first_party_id=1001, second_party_id=1002
        )
        # when
        result = obj.eve_entity_ids()
        # then
        expected = {1001, 1002}
        self.assertSetEqual(result, expected)


class TestCharacterWalletJournalEntry(NoSocketsTestCase):
    def test_match_context_type_id(self):
        self.assertEqual(
            CharacterWalletJournalEntry.match_context_type_id("character_id"),
            CharacterWalletJournalEntry.CONTEXT_ID_TYPE_CHARACTER_ID,
        )
        self.assertEqual(
            CharacterWalletJournalEntry.match_context_type_id("contract_id"),
            CharacterWalletJournalEntry.CONTEXT_ID_TYPE_CONTRACT_ID,
        )
        self.assertEqual(
            CharacterWalletJournalEntry.match_context_type_id(None),
            CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
        )



# flake8: noqa
"""Script for generating invalid locations to test fix for issue #153."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
from django.db.models import Max
from eveuniverse.models import EveEntity, EveType

from allianceauth.eveonline.models import EveCharacter

from memberaudit.models import Character, Location
from memberaudit.tests.testdata.constants import EveTypeId
from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_asset,
    create_character_contract_courier,
    create_character_jump_clone,
    create_character_jump_clone_implant,
    create_character_location,
    create_character_wallet_transaction,
    create_location,
)

EVE_CHARACTER_ID = 91224790  # CCP Seagull


def max_location_id():
    max_location_id = Location.objects.aggregate(Max("id"))["id__max"]
    return max_location_id


locations = Location.objects.filter(
    eve_type__isnull=False, eve_solar_system__isnull=False
)
valid_location_1 = locations.first()
valid_location_2 = locations.last()
if not valid_location_1 or valid_location_1 == valid_location_2:
    raise RuntimeError("Need two valid locations to exist")

character_entity = EveEntity.objects.filter(
    category=EveEntity.CATEGORY_CHARACTER
).first()
if not character_entity:
    raise RuntimeError("Need a character in eve entities to exist")

corporation_entity = EveEntity.objects.filter(
    category=EveEntity.CATEGORY_CORPORATION
).first()
if not corporation_entity:
    raise RuntimeError("Need a corporation in eve entities to exist")

try:
    eve_character = EveCharacter.objects.get(character_id=EVE_CHARACTER_ID)
except EveCharacter.DoesNotExist:
    eve_character = EveCharacter.objects.create_character(character_id=EVE_CHARACTER_ID)

Character.objects.filter(eve_character=eve_character).delete()
character = create_character(eve_character, is_disabled=True)

item_eve_type, _ = EveType.objects.get_or_create_esi(id=EveTypeId.VELDSPAR)
implant_eve_type, _ = EveType.objects.get_or_create_esi(
    id=EveTypeId.HIGH_GRADE_SNAKE_ALPHA
)

# create assets and invalid location
asset_1 = create_character_asset(
    character=character,
    eve_type=item_eve_type,
    location=valid_location_1,
    item_id=max_location_id() + 1,
)
invalid_location_1 = create_location(id=asset_1.item_id)
asset_2 = create_character_asset(
    character=character,
    eve_type=item_eve_type,
    location=valid_location_2,
    item_id=max_location_id() + 1,
)
invalid_location_2 = create_location(id=asset_2.item_id)

# create character objects
for location in [valid_location_1, invalid_location_1]:
    create_character_asset(
        character=character, eve_type=item_eve_type, location=location
    )
    jump_clone = create_character_jump_clone(character=character, location=location)
    create_character_jump_clone_implant(
        jump_clone=jump_clone, eve_type=implant_eve_type
    )
    create_character_wallet_transaction(
        character=character,
        location=location,
        eve_type=item_eve_type,
        client=corporation_entity,
    )

create_character_location(character=character, location=invalid_location_1)
create_character_contract_courier(
    character=character,
    assignee=None,
    issuer=character_entity,
    issuer_corporation=corporation_entity,
    start_location=valid_location_1,
    end_location=valid_location_2,
)
create_character_contract_courier(
    character=character,
    assignee=None,
    issuer=character_entity,
    issuer_corporation=corporation_entity,
    start_location=invalid_location_1,
    end_location=invalid_location_2,
)
for section in [
    Character.UpdateSection.ASSETS,
    Character.UpdateSection.CONTRACTS,
    Character.UpdateSection.LOCATION,
    Character.UpdateSection.JUMP_CLONES,
    Character.UpdateSection.WALLET_TRANSACTIONS,
]:
    character.update_section_log_result(section=section, is_success=True)

print("DONE!")



"""Manager for Character model."""
# pylint: disable=missing-class-docstring

from copy import deepcopy
from math import floor
from typing import Set

from django.conf import settings as auth_settings
from django.contrib.auth.models import Permission, User
from django.db import models
from django.db.models import (
    Avg,
    Case,
    Count,
    ExpressionWrapper,
    F,
    Max,
    Min,
    Q,
    Value,
    When,
)

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger
from app_utils.caching import ObjectCacheMixin
from app_utils.django import users_with_permission
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, app_settings

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterQuerySet(models.QuerySet):
    def eve_character_ids(self) -> Set[int]:
        """Return EveCharacter IDs of all characters in this QuerySet."""
        return set(self.values_list("eve_character__character_id", flat=True))

    def owned_by_user(self, user: User) -> models.QuerySet:
        """Filter character owned by user."""
        return self.filter(eve_character__character_ownership__user__pk=user.pk)

    def annotate_total_update_status(self) -> models.QuerySet:
        """Add total_update_status annotations."""
        from memberaudit.models import Character

        enabled_sections = list(Character.UpdateSection.enabled_sections())
        num_sections_total = len(enabled_sections)
        qs = (
            self.annotate(
                num_sections_total=Count(
                    "update_status_set",
                    filter=Q(update_status_set__section__in=enabled_sections),
                )
            )
            .annotate(
                num_sections_ok=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__is_success=True,
                    ),
                )
            )
            .annotate(
                num_sections_failed=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__is_success=False,
                    ),
                )
            )
            .annotate(
                num_sections_token_error=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__has_token_error=True,
                    ),
                )
            )
            .annotate(
                total_update_status=Case(
                    When(
                        is_disabled=True,
                        then=Value(Character.TotalUpdateStatus.DISABLED.value),
                    ),
                    When(
                        num_sections_token_error=1,
                        then=Value(Character.TotalUpdateStatus.LIMITED_TOKEN.value),
                    ),
                    When(
                        num_sections_failed__gt=0,
                        then=Value(Character.TotalUpdateStatus.ERROR.value),
                    ),
                    When(
                        num_sections_ok=num_sections_total,
                        then=Value(Character.TotalUpdateStatus.OK.value),
                    ),
                    When(
                        num_sections_total__lt=num_sections_total,
                        then=Value(Character.TotalUpdateStatus.INCOMPLETE.value),
                    ),
                    default=Value(Character.TotalUpdateStatus.IN_PROGRESS.value),
                )
            )
        )
        return qs

    def disable_characters_with_no_owner(self) -> int:
        """Disable characters which have no owner. Return count of disabled characters."""
        orphaned_characters = self.filter(
            eve_character__character_ownership__isnull=True, is_disabled=False
        )
        if orphaned_characters.exists():
            orphans = list(
                orphaned_characters.values_list(
                    "eve_character__character_name", flat=True
                ).order_by("eve_character__character_name")
            )
            orphaned_characters.update(is_disabled=True)
            logger.info(
                "Disabled %d characters which do not belong to a user: %s",
                len(orphans),
                ", ".join(orphans),
            )
            return len(orphans)

        return 0


class CharacterManagerBase(ObjectCacheMixin, models.Manager):
    def characters_of_user_to_register_count(self, user: User) -> int:
        """Return count of a users's characters known to Auth,
        which needs to be (re-)registered.
        """
        unregistered = CharacterOwnership.objects.filter(
            user=user, character__memberaudit_character__isnull=True
        ).count()
        enabled_sections = list(self.model.UpdateSection.enabled_sections())
        token_errors = (
            self.filter(eve_character__character_ownership__user=user)
            .filter(
                Q(
                    update_status_set__section__in=enabled_sections,
                    update_status_set__has_token_error=True,
                )
                | Q(is_disabled=True),
            )
            .distinct()
            .count()
        )
        return unregistered + token_errors

    def user_has_scope(self, user: User) -> models.QuerySet:
        """Return characters the given user has the scope permission to access."""
        if user.has_perm("memberaudit.view_everything"):
            return self.all()
        qs = self.filter(eve_character__character_ownership__user=user)
        if (
            user.has_perm("memberaudit.view_same_alliance")
            and user.profile.main_character.alliance_id
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__alliance_id=(
                    user.profile.main_character.alliance_id
                )
            )
        elif user.has_perm("memberaudit.view_same_corporation"):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__corporation_id=(
                    user.profile.main_character.corporation_id
                )
            )
        return qs

    def user_has_access(self, user: User) -> models.QuerySet:
        """Return characters the given user has permission to access
        via character viewer.
        """
        if user.has_perm("memberaudit.view_everything") and user.has_perm(
            "memberaudit.characters_access"
        ):
            return self.all()
        qs = self.filter(eve_character__character_ownership__user=user)
        if (
            user.has_perm("memberaudit.characters_access")
            and user.has_perm("memberaudit.view_same_alliance")
            and user.profile.main_character.alliance_id
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__alliance_id=(
                    user.profile.main_character.alliance_id
                )
            )
        elif user.has_perm("memberaudit.characters_access") and user.has_perm(
            "memberaudit.view_same_corporation"
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__corporation_id=(
                    user.profile.main_character.corporation_id
                )
            )
        if user.has_perm("memberaudit.view_shared_characters"):
            permission_to_share_characters = Permission.objects.select_related(
                "content_type"
            ).get(
                content_type__app_label=self.model._meta.app_label,
                codename="share_characters",
            )
            viewable_users = users_with_permission(permission_to_share_characters)
            qs |= self.filter(
                is_shared=True,
                eve_character__character_ownership__user__in=viewable_users,
            )
        return qs


CharacterManager = CharacterManagerBase.from_queryset(CharacterQuerySet)


class CharacterUpdateStatusQuerySet(models.QuerySet):
    def filter_enabled_sections(self) -> models.QuerySet:
        """Filter enabled sections."""
        from memberaudit.models import Character

        enabled_sections = list(Character.UpdateSection.enabled_sections())
        return self.filter(section__in=enabled_sections)


class CharacterUpdateStatusManagerBase(models.Manager):
    def statistics(self) -> dict:
        """Return detailed statistics about the last update run and the app."""

        from memberaudit.models import (
            Character,
            CharacterAsset,
            CharacterContact,
            CharacterContract,
            CharacterMail,
            SkillSet,
            SkillSetGroup,
        )

        all_characters_count = Character.objects.count()

        settings = self._fetch_settings()

        update_stats = self._calc_update_stats(all_characters_count, settings)

        return {
            "app_totals": {
                "users_count": User.objects.filter(
                    character_ownerships__character__memberaudit_character__isnull=False
                )
                .distinct()
                .count(),
                "all_characters_count": all_characters_count,
                "skill_set_groups_count": SkillSetGroup.objects.count(),
                "skill_sets_count": SkillSet.objects.count(),
                "assets_count": CharacterAsset.objects.count(),
                "mails_count": CharacterMail.objects.count(),
                "contacts_count": CharacterContact.objects.count(),
                "contracts_count": CharacterContract.objects.count(),
            },
            "settings": settings,
            "update_statistics": update_stats,
        }

    def _calc_update_stats(self, all_characters_count, settings):
        from memberaudit.models import Character

        update_stats = {}
        if self.count() > 0:
            qs_base, root_task_ids = self._calc_qs_base()
            # per ring
            for ring in range(1, 4):
                sections = Character.sections_in_ring(ring)

                # calc totals
                qs_ring = qs_base.filter(section__in=sections)
                self._calc_totals(
                    all_characters_count,
                    settings,
                    root_task_ids,
                    update_stats,
                    ring,
                    qs_ring,
                )

                # calc section stats
                for section in sections:
                    self._update_section(qs_base, section, update_stats, ring)

                self._update_character_ring_counts(
                    all_characters_count, update_stats, ring, sections, qs_ring
                )

        return update_stats

    def _update_character_ring_counts(
        self, all_characters_count, update_stats, ring, sections, qs
    ):
        from memberaudit.models import Character

        ring_characters_count = (
            Character.objects.filter(update_status_set__in=qs)
            .annotate(num_sections=Count("update_status_set__section"))
            .filter(num_sections=len(sections))
            .count()
        )
        update_stats[f"ring_{ring}"]["total"][
            "characters_count"
        ] = ring_characters_count
        update_stats[f"ring_{ring}"]["total"]["completed"] = (
            ring_characters_count == all_characters_count
        )

    def _calc_qs_base(self):
        from memberaudit.models import Character

        qs_base = self.filter(
            is_success=True,
            started_at__isnull=False,
            finished_at__isnull=False,
        ).exclude(root_task_id="", parent_task_id="")
        root_task_ids = {
            ring: self._root_task_id_or_none(
                qs_base.filter(section__in=Character.sections_in_ring(ring))
                .order_by("-finished_at")
                .first()
            )
            for ring in range(1, 4)
        }
        duration_expression = ExpressionWrapper(
            F("finished_at") - F("started_at"),
            output_field=models.fields.DurationField(),
        )
        qs_base = qs_base.filter(root_task_id__in=root_task_ids.values()).annotate(
            duration=duration_expression
        )
        return qs_base, root_task_ids

    def _calc_totals(
        self, all_characters_count, settings, root_task_ids, update_stats, ring, qs
    ):  # pylint: disable=too-many-locals
        try:
            first = qs.order_by("started_at").first()
            last = qs.order_by("finished_at").last()
            started_at = first.started_at
            finished_at = last.finished_at
            duration = round((finished_at - started_at).total_seconds(), 1)
        except (KeyError, AttributeError):
            first = None
            last = None
            duration = None
            started_at = None
            finished_at = None

        available_time = (
            settings[f"MEMBERAUDIT_UPDATE_STALE_RING_{ring}"]
            - settings["MEMBERAUDIT_UPDATE_STALE_OFFSET"]
        ) * 60
        throughput = floor(all_characters_count / duration * 3600) if duration else None
        within_boundaries = duration < available_time if duration else None
        update_stats[f"ring_{ring}"] = {
            "total": {
                "duration": duration,
                "started_at": started_at,
                "finished_at": finished_at,
                "root_task_id": root_task_ids.get(ring),
                "throughput_est": throughput,
                "available_time": available_time,
                "within_available_time": within_boundaries,
            },
            "max": {},
            "sections": {},
        }

        # add longest running section w/ character
        obj = qs.order_by("-duration").first()
        update_stats[f"ring_{ring}"]["max"] = self._info_from_obj(obj)

        # add first and last section
        update_stats[f"ring_{ring}"]["first"] = self._info_from_obj(first)
        update_stats[f"ring_{ring}"]["last"] = self._info_from_obj(last)

        return first, last

    def _update_section(self, qs_base, section, update_stats, ring):
        try:
            section_max = round(
                qs_base.filter(section=section)
                .aggregate(Max("duration"))["duration__max"]
                .total_seconds(),
                1,
            )
            section_avg = round(
                qs_base.filter(section=section)
                .aggregate(Avg("duration"))["duration__avg"]
                .total_seconds(),
                1,
            )
            section_min = round(
                qs_base.filter(section=section)
                .aggregate(Min("duration"))["duration__min"]
                .total_seconds(),
                1,
            )
        except (KeyError, AttributeError):
            section_max = (None,)
            section_avg = None
            section_min = None

        update_stats[f"ring_{ring}"]["sections"].update(
            {
                str(section): {
                    "max": section_max,
                    "avg": section_avg,
                    "min": section_min,
                }
            }
        )

    def _fetch_settings(self):
        settings = {
            name: value
            for name, value in vars(app_settings).items()
            if name.startswith("MEMBERAUDIT_")
        }
        schedule = deepcopy(auth_settings.CELERYBEAT_SCHEDULE)
        for name, details in schedule.items():
            for key, value in details.items():
                if key == "schedule":
                    schedule[name][key] = str(value)

        settings["CELERYBEAT_SCHEDULE"] = schedule
        return settings

    @staticmethod
    def _root_task_id_or_none(obj):
        try:
            return obj.root_task_id
        except AttributeError:
            return None

    @staticmethod
    def _info_from_obj(obj) -> dict:
        try:
            section_name = str(obj.section)
            character_name = str(obj.character)
            duration = round(obj.duration.total_seconds(), 1)
        except (KeyError, AttributeError):
            section_name = None
            character_name = None
            duration = None

        return {
            "section": section_name,
            "character": character_name,
            "duration": duration,
        }


CharacterUpdateStatusManager = CharacterUpdateStatusManagerBase.from_queryset(
    CharacterUpdateStatusQuerySet
)



from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from eveuniverse.models import EveType

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import (
    NoSocketsTestCase,
    create_authgroup,
    create_user_from_evecharacter,
    queryset_pks,
)

from memberaudit.models import (
    EveShipType,
    EveSkillType,
    General,
    Location,
    MailEntity,
    SkillSet,
    SkillSetGroup,
    SkillSetSkill,
)
from memberaudit.tests.testdata.constants import EveSolarSystemId
from memberaudit.tests.testdata.factories import (
    create_compliance_group_designation,
    create_location_eve_solar_system,
    create_mail_entity_from_eve_entity,
    create_mailing_list,
    create_skill_set,
    create_skill_set_skill,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    add_auth_character_to_user,
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    permissions_for_model,
)

MODELS_PATH = "memberaudit.models"
MANAGERS_PATH = "memberaudit.managers"
TASKS_PATH = "memberaudit.tasks"


class TestMailEntity(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_entities()

    def test_str(self):
        obj = create_mail_entity_from_eve_entity(1001)
        self.assertEqual(str(obj), "Bruce Wayne")

    def test_repr(self):
        obj = create_mail_entity_from_eve_entity(1001)
        self.assertEqual(
            repr(obj), "MailEntity(id=1001, category=CH, name='Bruce Wayne')"
        )

    def test_eve_entity_categories(self):
        obj = create_mail_entity_from_eve_entity(1001)
        self.assertSetEqual(
            obj.eve_entity_categories,
            {
                MailEntity.Category.ALLIANCE,
                MailEntity.Category.CHARACTER,
                MailEntity.Category.CORPORATION,
            },
        )

    def test_name_plus_1(self):
        obj = create_mail_entity_from_eve_entity(1001)
        self.assertEqual(obj.name_plus, "Bruce Wayne")

    def test_name_plus_2(self):
        obj = MailEntity.objects.create(id=42, category=MailEntity.Category.ALLIANCE)
        self.assertEqual(obj.name_plus, "Alliance #42")

    def test_need_to_specify_category(self):
        with self.assertRaises(ValidationError):
            MailEntity.objects.create(id=1)

    def test_url_1(self):
        obj = create_mail_entity_from_eve_entity(3001)
        self.assertIn("dotlan", obj.external_url())

    def test_url_2(self):
        obj = create_mail_entity_from_eve_entity(2001)
        self.assertIn("dotlan", obj.external_url())

    def test_url_3(self):
        obj = create_mail_entity_from_eve_entity(1001)
        self.assertIn("evewho", obj.external_url())

    def test_url_4(self):
        obj = create_mailing_list()
        self.assertEqual(obj.external_url(), "")

    def test_url_5(self):
        obj = MailEntity.objects.create(id=9887, category=MailEntity.Category.ALLIANCE)
        self.assertEqual(obj.external_url(), "")

    def test_url_6(self):
        obj = MailEntity.objects.create(
            id=9887, category=MailEntity.Category.CORPORATION
        )
        self.assertEqual(obj.external_url(), "")


class TestGeneralOther(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_entities()

    def test_should_return_compliant_users_only(self):
        # given
        # compliant user both chars registered
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        add_memberaudit_character_to_user(user_compliant, 1101)
        # non-compliant user one char not registered
        user_non_compliant_1, _ = create_user_from_evecharacter(
            1002, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1002)
        add_auth_character_to_user(user_non_compliant_1, 1102)
        # non-compliant user with char registered, but missing permission
        user_non_compliant_2, _ = create_user_from_evecharacter(1003)
        add_memberaudit_character_to_user(user_non_compliant_2, 1003)
        # when
        result = General.compliant_users()
        # then
        self.assertQuerysetEqual(result, User.objects.filter(pk=user_compliant.pk))

    def test_should_add_group_to_compliant_users(self):
        # given
        group = create_authgroup(internal=True)
        user_compliant, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user_compliant, 1001)
        user_non_compliant, _ = create_user_from_evecharacter(
            1002, permissions=["memberaudit.basic_access"]
        )
        # when
        General.add_compliant_users_to_group(group)
        # then
        self.assertIn(group, user_compliant.groups.all())
        self.assertNotIn(group, user_non_compliant.groups.all())


class TestGeneralUserHasAccess(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        character_1002 = create_memberaudit_character(1002)
        cls.user_1002 = character_1002.eve_character.character_ownership.user
        character_1003 = create_memberaudit_character(1003)
        cls.user_1003 = character_1003.eve_character.character_ownership.user
        character_1101 = create_memberaudit_character(1101)
        cls.user_1101 = character_1101.eve_character.character_ownership.user
        cls.user_dummy = AuthUtils.create_user("No-access-to-Member-Audit")

    def setUp(self) -> None:
        character_1001 = create_memberaudit_character(1001)
        self.user_1001 = character_1001.eve_character.character_ownership.user

    def test_should_see_own_user_only(self):
        # when
        result = General.accessible_users(user=self.user_1001)
        # then
        self.assertSetEqual(queryset_pks(result), {self.user_1001.pk})

    def test_should_see_all_memberaudit_users(self):
        # given
        self.user_1001 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_everything", self.user_1001
        )
        # when
        result = General.accessible_users(user=self.user_1001)
        # then
        self.assertSetEqual(
            queryset_pks(result),
            {
                self.user_1001.pk,
                self.user_1002.pk,
                self.user_1003.pk,
                self.user_1101.pk,
            },
        )

    def test_should_see_own_alliance_only(self):
        # given
        self.user_1001 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", self.user_1001
        )
        # when
        result = General.accessible_users(user=self.user_1001)
        # then
        self.assertSetEqual(
            queryset_pks(result),
            {self.user_1001.pk, self.user_1002.pk, self.user_1003.pk},
        )

    def test_should_see_own_corporation_only(self):
        # given
        self.user_1001 = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", self.user_1001
        )
        # when
        result = General.accessible_users(user=self.user_1001)
        # then
        self.assertSetEqual(
            queryset_pks(result), {self.user_1001.pk, self.user_1002.pk}
        )


class TestLocation(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.location_jita = create_location_eve_solar_system(id=EveSolarSystemId.JITA)
        cls.location_asset_safety: Location = Location.objects.get_or_create_esi(
            id=2004, token=None
        )[0]
        cls.location_jita_44: Location = Location.objects.get(id=60003760)
        cls.location_structure_1: Location = Location.objects.get(id=1_000_000_000_001)
        cls.location_empty = Location(id=1)
        cls.location_unknown, _ = Location.objects.get_or_create_esi(
            id=Location.LOCATION_UNKNOWN_ID, token=None
        )

    def test_str(self):
        self.assertEqual(
            str(self.location_structure_1), "Amamake - Test Structure Alpha"
        )

    def test_checks_with_solar_system(self):
        location = self.location_jita
        self.assertTrue(location.is_solar_system)
        self.assertFalse(location.is_station)
        self.assertFalse(location.is_structure)
        self.assertFalse(Location.is_asset_safety_id(location.id))
        self.assertFalse(location.is_empty)

    def test_checks_with_station(self):
        location = self.location_jita_44
        self.assertFalse(location.is_solar_system)
        self.assertTrue(location.is_station)
        self.assertFalse(location.is_structure)
        self.assertFalse(Location.is_asset_safety_id(location.id))
        self.assertFalse(location.is_empty)

    def test_checks_with_structure(self):
        location = self.location_structure_1
        self.assertFalse(location.is_solar_system)
        self.assertFalse(location.is_station)
        self.assertTrue(location.is_structure)
        self.assertFalse(Location.is_asset_safety_id(location.id))
        self.assertFalse(location.is_empty)

    def test_checks_with_asset_safety(self):
        location = self.location_asset_safety
        self.assertFalse(location.is_solar_system)
        self.assertFalse(location.is_station)
        self.assertFalse(location.is_structure)
        self.assertTrue(Location.is_asset_safety_id(location.id))
        self.assertFalse(location.is_empty)

    def test_checks_with_empty_location(self):
        location = self.location_empty
        self.assertFalse(location.is_solar_system)
        self.assertFalse(location.is_station)
        self.assertFalse(location.is_structure)
        self.assertFalse(Location.is_asset_safety_id(location.id))
        self.assertTrue(location.is_empty)

    def test_solar_system_url(self):
        obj_1 = self.location_structure_1
        obj_2 = Location.objects.create(id=1_000_000_000_999)

        self.assertIn("Amamake", obj_1.solar_system_url)
        self.assertEqual("", obj_2.solar_system_url)

    def test_name_plus_for_structure(self):
        self.assertEqual(
            self.location_structure_1.name_plus, "Amamake - Test Structure Alpha"
        )

    def test_name_plus_for_station(self):
        self.assertEqual(
            self.location_jita_44.name_plus,
            "Jita IV - Moon 4 - Caldari Navy Assembly Plant",
        )

    def test_name_plus_for_solar_system(self):
        self.assertEqual(self.location_jita.name_plus, "Jita")

    def test_name_plus_for_asset_safety(self):
        self.assertEqual(self.location_asset_safety.name_plus, "ASSET SAFETY")

    def test_should_return_correct_asset_location_type(self):
        cases = [
            ("station", self.location_jita_44, "station"),
            ("structure", self.location_structure_1, "item"),
            ("solar system", self.location_jita, "solar_system"),
            ("unknown placeholder", self.location_unknown, "solar_system"),
            ("empty_location", self.location_empty, "other"),
        ]
        for name, location, expected in cases:
            with self.subTest(name=name):
                self.assertEqual(location.asset_location_type(), expected)


class TestComplianceGroupDesignation(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_entities()

    def test_should_ensure_new_compliance_groups_are_internal(self):
        # given
        group = create_authgroup(internal=False)
        # when
        create_compliance_group_designation(group)
        # then
        group.refresh_from_db()
        self.assertTrue(group.authgroup.internal)


class TestSkillSet(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        cls.user = AuthUtils.create_user("Bruce Wayne")

    def test_should_clone_a_skill_set(self):
        # given
        obj_1 = create_skill_set()
        gunnery_skill = EveType.objects.get(name="Gunnery")
        skill_1 = create_skill_set_skill(
            obj_1, eve_type=gunnery_skill, required_level=3, recommended_level=5
        )
        # when
        obj_2 = obj_1.clone(user=self.user)
        # then
        self.assertNotEqual(obj_2.pk, obj_1.pk)
        self.assertEqual(obj_2.description, obj_1.description)
        self.assertEqual(obj_2.is_visible, obj_1.is_visible)
        self.assertNotEqual(obj_2.last_modified_at, obj_1.last_modified_at)
        self.assertEqual(obj_2.last_modified_by, self.user)
        self.assertEqual(obj_2.ship_type, obj_1.ship_type)

        skill_2: SkillSetSkill = obj_2.skills.first()
        self.assertNotEqual(skill_2.pk, skill_1.pk)
        self.assertEqual(skill_2.eve_type, skill_1.eve_type)
        self.assertEqual(skill_2.required_level, skill_1.required_level)
        self.assertEqual(skill_2.recommended_level, skill_1.recommended_level)


class TestPermissions(NoSocketsTestCase):
    def test_should_have_default_permissions_for_skill_set_models(self):
        for model_class in [
            EveSkillType,
            EveShipType,
            SkillSet,
            SkillSetGroup,
            SkillSetSkill,
        ]:
            with self.subTest(model=model_class.__name__):
                # when/then
                self.assertTrue(permissions_for_model(model_class).exists())



from unittest.mock import patch

from bravado.exception import HTTPNotFound

from eveuniverse.models import EveEntity, EveType

from app_utils.esi_testing import BravadoResponseStub
from app_utils.testing import NoSocketsTestCase

from memberaudit.core.eft_parser import (
    MissingSectionsError,
    MissingTitleError,
    UnknownShipTypeError,
    _EftItem,
    _EftSection,
    _EftSlotType,
    _EftTextItem,
    _EftTextSection,
    _EveTypes,
    create_fitting_from_eft,
)
from memberaudit.tests.testdata.factories import create_fitting_text
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "memberaudit.core.eft_parser"


class TestEftParser(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()

    def test_should_create_fitting(self):
        # given
        fitting_text = create_fitting_text("fitting_tristan.txt")
        # when
        fitting, errors = create_fitting_from_eft(fitting_text)
        # then
        self.assertListEqual(errors, [])
        self.assertEqual(fitting.name, "Tristan - Standard Kite (cap stable)")
        self.assertEqual(fitting.ship_type.name, "Tristan")
        self.assertEqual(
            fitting.low_slots[0].module_type.name, "Nanofiber Internal Structure II"
        )
        self.assertEqual(
            fitting.low_slots[1].module_type.name, "Drone Damage Amplifier II"
        )
        self.assertEqual(
            fitting.low_slots[2].module_type.name, "Drone Damage Amplifier II"
        )

        self.assertEqual(
            fitting.medium_slots[0].module_type.name,
            "5MN Cold-Gas Enduring Microwarpdrive",
        )
        self.assertEqual(
            fitting.medium_slots[1].module_type.name,
            "Medium Azeotropic Restrained Shield Extender",
        )
        self.assertEqual(fitting.medium_slots[2].module_type.name, "Warp Disruptor II")
        self.assertTrue(fitting.medium_slots[2].is_offline)

        self.assertEqual(
            fitting.high_slots[0].module_type.name, "125mm Gatling AutoCannon II"
        )
        self.assertEqual(fitting.high_slots[0].charge_type.name, "EMP S")
        self.assertEqual(
            fitting.high_slots[1].module_type.name, "125mm Gatling AutoCannon II"
        )
        self.assertEqual(fitting.high_slots[1].charge_type.name, "EMP S")
        self.assertTrue(fitting.high_slots[2].is_empty)

        self.assertEqual(
            fitting.rig_slots[0].module_type.name, "Small Capacitor Control Circuit I"
        )
        self.assertEqual(
            fitting.rig_slots[1].module_type.name, "Small Polycarbon Engine Housing I"
        )
        self.assertEqual(
            fitting.rig_slots[2].module_type.name, "Small EM Shield Reinforcer I"
        )

        self.assertEqual(fitting.drone_bay[0].item_type.name, "Acolyte II")
        self.assertEqual(fitting.drone_bay[0].quantity, 5)
        self.assertEqual(fitting.drone_bay[1].item_type.name, "Warrior II")
        self.assertEqual(fitting.drone_bay[1].quantity, 3)

    def test_should_raise_error_when_title_is_missing(self):
        # given
        fitting_text = create_fitting_text("fitting_tristan_no_title.txt")
        # when
        with self.assertRaises(MissingTitleError):
            create_fitting_from_eft(fitting_text)

    def test_should_raise_error_when_text_is_empty(self):
        # when
        with self.assertRaises(MissingSectionsError):
            create_fitting_from_eft("")

    def test_should_raise_error_when_ship_type_is_invalid(self):
        # given
        fitting_text = create_fitting_text("fitting_unknown_ship_type.txt")
        # when
        with self.assertRaises(UnknownShipTypeError):
            with patch(MODULE_PATH + "._load_eve_types") as m:
                m.return_value = _EveTypes(), {"Unknown"}
                create_fitting_from_eft(fitting_text)

    def test_should_report_unknown_types(self):
        # given
        fitting_text = create_fitting_text("fitting_tristan_unknown_types.txt")
        # when
        with patch(MODULE_PATH + ".EveEntity.objects.fetch_by_names_esi") as mock:
            mock.return_value = EveEntity.objects.none()
            fitting, errors = create_fitting_from_eft(fitting_text)
        # then
        self.assertIn("Unknown Type Alpha", ", ".join(errors))
        self.assertIn("Unknown Type Bravo", ", ".join(errors))
        self.assertEqual(
            fitting.low_slots[0].module_type.name, "Nanofiber Internal Structure II"
        )

    def test_should_handle_eve_format_with_missing_high_slots(self):
        # given
        fitting_text = create_fitting_text("fitting_eve_tristian.txt")
        # when
        fitting, errors = create_fitting_from_eft(fitting_text)
        # then
        self.assertListEqual(errors, [])
        self.assertListEqual(fitting.high_slots, [])
        self.assertListEqual(
            section_to_eft(fitting.low_slots),
            [
                "Nanofiber Internal Structure II",
                "Drone Damage Amplifier II",
                "Drone Damage Amplifier II",
            ],
        )
        self.assertListEqual(
            section_to_eft(fitting.medium_slots),
            [
                "5MN Cold-Gas Enduring Microwarpdrive",
                "Medium Azeotropic Restrained Shield Extender",
                "Warp Disruptor II",
            ],
        )
        self.assertListEqual(
            section_to_eft(fitting.rig_slots),
            [
                "Small Capacitor Control Circuit I",
                "Small Polycarbon Engine Housing I",
                "Small EM Shield Reinforcer I",
            ],
        )
        self.assertListEqual(
            section_to_eft(fitting.drone_bay),
            ["Warrior II x3", "Acolyte II x5"],
        )
        self.assertListEqual(
            section_to_eft(fitting.cargo_bay),
            ["EMP S x400"],
        )

    def test_should_handle_eve_format_with_missing_high_and_low_slots(self):
        # given
        fitting_text = create_fitting_text("fitting_eve_tristian_2.txt")
        # when
        fitting, errors = create_fitting_from_eft(fitting_text)
        # then
        self.assertListEqual(errors, [])
        self.assertListEqual(fitting.high_slots, [])
        self.assertListEqual(fitting.low_slots, [])
        self.assertListEqual(
            section_to_eft(fitting.medium_slots),
            [
                "5MN Cold-Gas Enduring Microwarpdrive",
                "Medium Azeotropic Restrained Shield Extender",
                "Warp Disruptor II",
            ],
        )
        self.assertListEqual(
            section_to_eft(fitting.rig_slots),
            [
                "Small Capacitor Control Circuit I",
                "Small Polycarbon Engine Housing I",
                "Small EM Shield Reinforcer I",
            ],
        )
        self.assertListEqual(
            section_to_eft(fitting.drone_bay),
            ["Warrior II x3", "Acolyte II x5"],
        )
        self.assertListEqual(
            section_to_eft(fitting.cargo_bay),
            ["EMP S x400"],
        )


def section_to_eft(items) -> list:
    return [obj.to_eft() for obj in items]


class TestEveTypes(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()

    def test_should_create_from_names(self):
        # given
        drones = EveType.objects.get(name="Drones")
        gunnery = EveType.objects.get(name="Gunnery")
        # when
        eve_types, _ = _EveTypes.create_from_names(["Drones", "Gunnery"])
        # then
        self.assertEqual(eve_types.from_name("Drones"), drones)
        self.assertEqual(eve_types.from_name("Gunnery"), gunnery)

    def test_should_try_to_fetch_unknown_types_from_esi(self):
        # given
        unknown_type = EveType(id=99, name="Unknown-Type")
        # when
        with patch(
            MODULE_PATH + ".EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch_by_names_esi, patch(
            MODULE_PATH + ".EveType.objects.get_or_create_esi"
        ) as mock_get_or_create_esi:
            mock_fetch_by_names_esi.return_value.filter.return_value.values_list.return_value = [
                99
            ]
            mock_get_or_create_esi.return_value = (unknown_type, False)
            eve_types, _ = _EveTypes.create_from_names(["Unknown-Type"])
        # then
        self.assertEqual(eve_types.from_name("Unknown-Type"), unknown_type)

    def test_should_handle_type_not_found(self):
        # given
        http404 = HTTPNotFound(
            BravadoResponseStub(status_code=404), message="Test exception"
        )
        # when
        with patch(
            MODULE_PATH + ".EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch_by_names_esi, patch(
            MODULE_PATH + ".EveType.objects.get_or_create_esi"
        ) as mock_get_or_create_esi:
            mock_fetch_by_names_esi.return_value.filter.return_value.values_list.return_value = [
                99
            ]
            mock_get_or_create_esi.side_effect = http404
            eve_types, unknown_types = _EveTypes.create_from_names(["Unknown-Type"])
        # then
        self.assertIsNone(eve_types.from_name("Unknown-Type"))
        self.assertIn("Unknown-Type", unknown_types)

    def test_should_handle_unknown_types(self):
        # given
        drones = EveType.objects.get(name="Drones")
        # when
        with patch(
            MODULE_PATH + ".EveEntity.objects.fetch_by_names_esi"
        ) as mock_fetch_by_names_esi:
            mock_fetch_by_names_esi.return_value.filter.return_value.values_list.return_value = (
                []
            )
            eve_types, unknown_types = _EveTypes.create_from_names(
                ["Drones", "Unknown-Type"]
            )
        # then
        self.assertEqual(eve_types.from_name("Drones"), drones)
        self.assertIsNone(eve_types.from_name("Unknown-Type"))
        self.assertIn("Unknown-Type", unknown_types)

    def test_should_raise_error_when_not_passing_iterable(self):
        # when/then
        with self.assertRaises(TypeError):
            _EveTypes.create_from_names("not an allowed iterable")


class TestEftTextItem(NoSocketsTestCase):
    def test_should_create_slot_module(self):
        #  when
        item = _EftTextItem.create_from_line("Warp Disruptor II")
        # then
        self.assertEqual(item, _EftTextItem(item_type="Warp Disruptor II"))

    def test_should_create_slot_module_with_charge(self):
        #  when
        item = _EftTextItem.create_from_line("125mm Gatling AutoCannon II, EMP S")
        # then
        self.assertEqual(
            item,
            _EftTextItem(item_type="125mm Gatling AutoCannon II", charge_type="EMP S"),
        )

    def test_should_create_slot_module_offline(self):
        #  when
        item = _EftTextItem.create_from_line("Warp Disruptor II /OFFLINE")
        # then
        self.assertEqual(
            item, _EftTextItem(item_type="Warp Disruptor II", is_offline=True)
        )

    def test_should_create_slot_empty_high_slot(self):
        #  when
        item = _EftTextItem.create_from_line("[Empty High slot]")
        # then
        expected = _EftTextItem(slot_type=_EftSlotType.HIGH_SLOT)
        self.assertEqual(item, expected)

    def test_should_create_slot_empty_sub_system_slot(self):
        #  when
        item = _EftTextItem.create_from_line("[Empty Subsystem slot]")
        # then
        expected = _EftTextItem(slot_type=_EftSlotType.SUBSYSTEM_SLOT)
        self.assertEqual(item, expected)

    def test_should_create_non_slot_item(self):
        #  when
        item = _EftTextItem.create_from_line("Acolyte II x5")
        # then
        self.assertEqual(item, _EftTextItem(item_type="Acolyte II", quantity=5))

    def test_should_return_type_names_1(self):
        #  when
        item = _EftTextItem.create_from_line("125mm Gatling AutoCannon II")
        # then
        self.assertEqual(item.type_names(), {"125mm Gatling AutoCannon II"})

    def test_should_return_type_names_2(self):
        #  when
        item = _EftTextItem.create_from_line("125mm Gatling AutoCannon II, EMP S")
        # then
        self.assertEqual(item.type_names(), {"125mm Gatling AutoCannon II", "EMP S"})

    def test_should_return_type_names_3(self):
        #  when
        item = _EftTextItem.create_from_line("Acolyte II x5")
        # then
        self.assertEqual(item.type_names(), {"Acolyte II"})


class TestEftTextSection(NoSocketsTestCase):
    def test_should_create_section(self):
        # given
        lines = ["Acolyte II x5", "Warrior II x5"]
        # when
        section = _EftTextSection.create_from_lines(lines)
        # then
        self.assertEqual(
            section.items[0], _EftTextItem(item_type="Acolyte II", quantity=5)
        )
        self.assertEqual(
            section.items[1], _EftTextItem(item_type="Warrior II", quantity=5)
        )

    def test_should_return_type_names(self):
        # given
        section = _EftTextSection(
            [
                _EftTextItem(item_type="Acolyte II", quantity=5),
                _EftTextItem(
                    item_type="125mm Gatling AutoCannon II", charge_type="EMP S"
                ),
                _EftTextItem(item_type="Warp Disruptor II"),
            ]
        )
        # when / then
        self.assertSetEqual(
            section.type_names(),
            {"Acolyte II", "125mm Gatling AutoCannon II", "EMP S", "Warp Disruptor II"},
        )


class TestEftItem(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()

    def test_should_create_from_text_item_1(self):
        # given
        text_item = _EftTextItem(item_type="Warp Disruptor II")
        eve_types, _ = _EveTypes.create_from_names(["Warp Disruptor II"])
        # when
        item = _EftItem.create_from_text_item(text_item=text_item, eve_types=eve_types)
        # then
        self.assertEqual(
            item, _EftItem(item_type=EveType.objects.get(name="Warp Disruptor II"))
        )

    def test_should_create_from_text_item_2(self):
        # given
        text_item = _EftTextItem(item_type="Warp Disruptor II", quantity=3)
        eve_types, _ = _EveTypes.create_from_names(["Warp Disruptor II"])
        # when
        item = _EftItem.create_from_text_item(text_item=text_item, eve_types=eve_types)
        # then
        self.assertEqual(
            item,
            _EftItem(
                item_type=EveType.objects.get(name="Warp Disruptor II"), quantity=3
            ),
        )

    def test_should_create_from_text_item_3(self):
        # given
        text_item = _EftTextItem(
            item_type="125mm Gatling AutoCannon II", charge_type="EMP S"
        )
        eve_types, _ = _EveTypes.create_from_names(
            ["125mm Gatling AutoCannon II", "EMP S"]
        )
        # when
        item = _EftItem.create_from_text_item(text_item=text_item, eve_types=eve_types)
        # then
        self.assertEqual(
            item,
            _EftItem(
                item_type=EveType.objects.get(name="125mm Gatling AutoCannon II"),
                charge_type=EveType.objects.get(name="EMP S"),
            ),
        )

    def test_should_be_slot(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="Warp Disruptor II"))
        # when/then
        self.assertTrue(item.is_slot)

    def test_should_not_be_slot(self):
        # given
        item = _EftItem(
            item_type=EveType.objects.get(name="Warp Disruptor II"), quantity=3
        )
        # when/then
        self.assertFalse(item.is_slot)

    def test_should_be_booster(self):
        # given
        item = _EftItem(
            item_type=EveType.objects.get(name="Agency 'Pyrolancea' DB5 Dose II")
        )
        # when/then
        self.assertTrue(item.is_booster())

    def test_should_be_cyber_implant(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="High-grade Snake Alpha"))
        # when/then
        self.assertTrue(item.is_cyber_implant())

    def test_should_be_drone(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="Acolyte II"))
        # when/then
        self.assertTrue(item.is_drone())

    def test_should_be_fighter(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="Firbolg I"))
        # when/then
        self.assertTrue(item.is_fighter())

    def test_should_be_subsystem(self):
        # given
        item = _EftItem(
            item_type=EveType.objects.get(
                name="Tengu Core - Augmented Graviton Reactor"
            )
        )
        # when/then
        self.assertTrue(item.is_subsystem())

    def test_should_be_high_slot(self):
        # given
        item = _EftItem(
            item_type=EveType.objects.get(name="125mm Gatling AutoCannon II")
        )
        # when/then
        self.assertTrue(item.is_high_slot())
        self.assertFalse(item.is_med_slot())
        self.assertFalse(item.is_low_slot())
        self.assertFalse(item.is_rig_slot())

    def test_should_be_med_slot(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="Warp Disruptor II"))
        # when/then
        self.assertFalse(item.is_high_slot())
        self.assertTrue(item.is_med_slot())
        self.assertFalse(item.is_low_slot())
        self.assertFalse(item.is_rig_slot())

    def test_should_be_low_slot(self):
        # given
        item = _EftItem(item_type=EveType.objects.get(name="Drone Damage Amplifier II"))
        # when/then
        self.assertFalse(item.is_high_slot())
        self.assertFalse(item.is_med_slot())
        self.assertTrue(item.is_low_slot())
        self.assertFalse(item.is_rig_slot())

    def test_should_be_rig_slot(self):
        # given
        item = _EftItem(
            item_type=EveType.objects.get(name="Small EM Shield Reinforcer I")
        )
        # when/then
        self.assertFalse(item.is_high_slot())
        self.assertFalse(item.is_med_slot())
        self.assertFalse(item.is_low_slot())
        self.assertTrue(item.is_rig_slot())

    def test_should_handle_empty_item(self):
        # given
        empty_item = _EftItem()
        # when/then
        self.assertTrue(empty_item.is_empty)
        self.assertFalse(empty_item.is_booster())
        self.assertFalse(empty_item.is_cyber_implant())
        self.assertFalse(empty_item.is_drone())
        self.assertFalse(empty_item.is_fighter())
        self.assertFalse(empty_item.is_high_slot())
        self.assertFalse(empty_item.is_med_slot())
        self.assertFalse(empty_item.is_low_slot())
        self.assertFalse(empty_item.is_rig_slot())
        self.assertFalse(empty_item.is_subsystem())


class TestEftSection(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()

    def test_should_be_boosters(self):
        # given
        section = _EftSection(
            [
                _EftItem(
                    item_type=EveType.objects.get(
                        name="Agency 'Pyrolancea' DB5 Dose II"
                    )
                )
            ]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.BOOSTERS)

    def test_should_be_implants(self):
        # given
        section = _EftSection(
            [_EftItem(item_type=EveType.objects.get(name="High-grade Snake Alpha"))]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.IMPLANTS)

    def test_should_be_drone_bay(self):
        # given
        section = _EftSection(
            [_EftItem(item_type=EveType.objects.get(name="Acolyte II"), quantity=5)]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.DRONES_BAY)

    def test_should_be_fighter_bay(self):
        # given
        section = _EftSection(
            [_EftItem(item_type=EveType.objects.get(name="Firbolg I"), quantity=5)]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.FIGHTER_BAY)

    def test_should_be_subsystem_slots(self):
        # given
        section = _EftSection(
            [
                _EftItem(
                    item_type=EveType.objects.get(
                        name="Tengu Core - Augmented Graviton Reactor"
                    )
                )
            ]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.SUBSYSTEM_SLOTS)

    def test_should_be_high_slots(self):
        # given
        section = _EftSection(
            [
                _EftItem(
                    item_type=EveType.objects.get(name="125mm Gatling AutoCannon II")
                )
            ]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.HIGH_SLOTS)

    def test_should_be_med_slots(self):
        # given
        section = _EftSection(
            [_EftItem(item_type=EveType.objects.get(name="Warp Disruptor II"))]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.MEDIUM_SLOTS)

    def test_should_be_low_slots(self):
        # given
        section = _EftSection(
            [_EftItem(item_type=EveType.objects.get(name="Drone Damage Amplifier II"))]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.LOW_SLOTS)

    def test_should_be_rig_slots(self):
        # given
        section = _EftSection(
            [
                _EftItem(
                    item_type=EveType.objects.get(name="Small EM Shield Reinforcer I")
                )
            ]
        )
        # when/then
        self.assertEqual(section.guess_category(), _EftSection.Category.RIG_SLOTS)

    def test_should_be_slots(self):
        # given
        section = _EftSection(
            [
                _EftItem(
                    item_type=EveType.objects.get(name="Small EM Shield Reinforcer I")
                )
            ]
        )
        # when/then
        self.assertTrue(section.is_slots)


class TestCreateFittingFromEft(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()

    def test_eft_parser_roundtrip_archon_normal(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_archon.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_eft_parser_rountrip_archon_max(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_archon_max.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_eft_parser_rountrip_tristan(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_tristan.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_eft_parser_rountrip_svipul_empty_slots_and_offline(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_svipul_2.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_eft_parser_rountrip_tengu(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_tengu.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # print(
        #     ", ".join(map(str, sorted(list([obj.id for obj in fitting.eve_types()]))))
        # )
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_eft_parser_rountrip_empty(self):
        # given
        self.maxDiff = None
        fitting_text_original = create_fitting_text("fitting_empty.txt")
        fitting, _ = create_fitting_from_eft(fitting_text_original)
        # when
        fitting_text_generated = fitting.to_eft()
        # then
        self.assertEqual(fitting_text_original, fitting_text_generated)

    def test_required_skills(self):
        # given
        fitting_text = create_fitting_text("fitting_tristan.txt")
        fitting, _ = create_fitting_from_eft(fitting_text)
        # when
        skills = fitting.required_skills()
        # then
        skills_str = sorted([str(skill) for skill in skills])
        self.assertListEqual(
            skills_str,
            [
                "Amarr Drone Specialization I",
                "Drones V",
                "Gallente Frigate I",
                "Gunnery II",
                "High Speed Maneuvering I",
                "Hull Upgrades II",
                "Light Drone Operation V",
                "Minmatar Drone Specialization I",
                "Propulsion Jamming II",
                "Shield Upgrades I",
                "Small Autocannon Specialization I",
                "Small Projectile Turret V",
                "Weapon Upgrades IV",
            ],
        )



"""Managers for character section models (1/3)."""
# pylint: disable=missing-class-docstring

from typing import Optional, Set

from django.db import models, transaction
from django.db.models import Case, ExpressionWrapper, F, Value, When
from esi.models import Token
from eveuniverse.models import EveEntity, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import (
    MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
    MEMBERAUDIT_DEVELOPER_MODE,
)
from memberaudit.decorators import fetch_token_for_character
from memberaudit.helpers import (
    data_retention_cutoff,
    eve_entity_ids_from_objs,
    store_debug_data_to_disk,
)
from memberaudit.providers import esi
from memberaudit.utils import (
    get_or_create_esi_or_none,
    get_or_create_or_none,
    get_or_none,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterAssetQuerySet(models.QuerySet):
    def item_ids(self) -> Set[int]:
        """Return item IDs for objects in this queryset."""
        return set(self.values_list("item_id", flat=True))

    def annotate_pricing(self) -> models.QuerySet:
        """Return qs with annotated price and total columns."""
        return (
            self.select_related("eve_type__market_price")
            .annotate(
                price=Case(
                    When(
                        is_blueprint_copy=True,
                        then=Value(None),
                    ),
                    default=F("eve_type__market_price__average_price"),
                )
            )
            .annotate(
                total=Case(
                    When(
                        is_blueprint_copy=True,
                        then=Value(None),
                    ),
                    default=ExpressionWrapper(
                        F("eve_type__market_price__average_price") * F("quantity"),
                        output_field=models.FloatField(),
                    ),
                )
            )
        )


class CharacterAssetManagerBase(models.Manager):
    def fetch_from_esi(self, character, force_update: bool = False) -> Optional[list]:
        """Fetch assets from ESI and preload related objects from ESI."""
        asset_list, changed = character.update_section_if_changed(
            section=character.UpdateSection.ASSETS,
            fetch_func=self._fetch_data_from_esi,
            store_func=None,
            force_update=force_update,
        )
        if changed:
            return asset_list

        return None

    @fetch_token_for_character("esi-assets.read_assets.v1")
    def _fetch_data_from_esi(self, character, token: Token) -> list:
        """Fetch character assets with names from ESI and return it."""
        logger.info("%s: Fetching assets from ESI", character)
        asset_list = esi.client.Assets.get_characters_character_id_assets(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        asset_data = {int(item["item_id"]): item for item in asset_list}

        logger.info("%s: Fetching asset names from ESI", character)
        names = []
        for asset_ids_chunk in chunks(list(asset_data.keys()), 999):
            names += esi.client.Assets.post_characters_character_id_assets_names(
                character_id=character.eve_character.character_id,
                token=token.valid_access_token(),
                item_ids=asset_ids_chunk,
            ).results()

        asset_names = {
            int(item["item_id"]): item["name"]
            for item in names
            if item["name"] != "None"
        }
        for item_id in asset_data.keys():
            asset_data[item_id]["name"] = asset_names.get(item_id, "")

        return list(asset_data.values())

    @fetch_token_for_character("esi-universe.read_structures.v1")
    def preload_objects_from_esi(
        self, character, token: Token, asset_list: list
    ) -> None:
        """Preloads objects needed to build the asset tree."""
        from memberaudit.models import Location

        if not asset_list:
            return

        logger.info("%s: Preloading objects for asset tree", character)
        required_ids = {item["type_id"] for item in asset_list if "type_id" in item}
        existing_ids = set(EveType.objects.values_list("id", flat=True))
        missing_ids = required_ids.difference(existing_ids)
        if missing_ids:
            logger.info(
                "%s: Loading %s missing types from ESI", character, len(missing_ids)
            )
            EveType.objects.bulk_get_or_create_esi(ids=list(missing_ids))

        asset_item_ids = {asset["item_id"] for asset in asset_list}
        incoming_location_ids = {
            item["location_id"]
            for item in asset_list
            if "location_id" in item and item["location_id"] not in asset_item_ids
        }
        Location.objects.create_missing_esi(
            location_ids=incoming_location_ids, token=token
        )


CharacterAssetManager = CharacterAssetManagerBase.from_queryset(CharacterAssetQuerySet)


class CharacterAttributesManager(models.Manager):
    def update_or_create_esi(self, character, force_update: bool = False):
        """Update or create attributes for a character from ESI."""
        character.update_section_if_changed(
            section=character.UpdateSection.ATTRIBUTES,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-skills.read_skills.v1")
    def _fetch_data_from_esi(self, character, token):
        logger.info("%s: Fetching attributes from ESI", character)
        attribute_data = esi.client.Skills.get_characters_character_id_attributes(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return attribute_data

    def _update_or_create_objs(self, character, attribute_data):
        self.update_or_create(
            character=character,
            defaults={
                "accrued_remap_cooldown_date": attribute_data.get(
                    "accrued_remap_cooldown_date"
                ),
                "last_remap_date": attribute_data.get("last_remap_date"),
                "bonus_remaps": attribute_data.get("bonus_remaps"),
                "charisma": attribute_data.get("charisma"),
                "intelligence": attribute_data.get("intelligence"),
                "memory": attribute_data.get("memory"),
                "perception": attribute_data.get("perception"),
                "willpower": attribute_data.get("willpower"),
            },
        )


class CharacterContactLabelManager(models.Manager):
    def update_or_create_esi(self, character, force_update: bool = False):
        """Update or create assets for a character from ESI."""

        character.update_section_if_changed(
            section=character.UpdateSection.CONTACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
            hash_num=2,
        )

    @fetch_token_for_character("esi-characters.read_contacts.v1")
    def _fetch_data_from_esi(self, character, token):
        logger.info("%s: Fetching contact labels from ESI", character)
        labels = esi.client.Contacts.get_characters_character_id_contacts_labels(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return labels

    @transaction.atomic()
    def _update_or_create_objs(self, character, labels):
        # TODO: replace with bulk methods to optimize
        if labels:
            incoming_ids = {label["label_id"] for label in labels}
        else:
            incoming_ids = set()
        existing_ids = set(
            self.filter(character=character).values_list("label_id", flat=True)
        )
        obsolete_ids = existing_ids.difference(incoming_ids)
        if obsolete_ids:
            logger.info("%s: Removing %s obsolete skills", character, len(obsolete_ids))
            self.filter(character=character, label_id__in=obsolete_ids).delete()
        if incoming_ids:
            logger.info("%s: Storing %s contact labels", character, len(incoming_ids))
            for label in labels:
                self.update_or_create(
                    character=character,
                    label_id=label.get("label_id"),
                    defaults={
                        "name": label.get("label_name"),
                    },
                )
        else:
            logger.info("%s: No contact labels", character)


class CharacterContactManager(models.Manager):
    def update_or_create_esi(self, character, force_update: bool = False):
        """Update or create assets for a character from ESI."""
        character.update_section_if_changed(
            section=character.UpdateSection.CONTACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_contacts.v1")
    def _fetch_data_from_esi(self, character, token):
        logger.info("%s: Fetching contacts from ESI", character)
        contacts_data = esi.client.Contacts.get_characters_character_id_contacts(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return contacts_data

    @transaction.atomic()
    def _update_or_create_objs(self, character, contacts_data) -> Set[int]:
        """Update or create new contact objects from provided data.

        Return EveEntity IDs in newly created contacts.
        """
        contacts_list = (
            {int(obj["contact_id"]): obj for obj in contacts_data}
            if contacts_data
            else {}
        )
        incoming_ids = set(contacts_list.keys())
        existing_ids = set(
            self.filter(character=character).values_list("eve_entity_id", flat=True)
        )
        obsolete_ids = existing_ids.difference(incoming_ids)
        if obsolete_ids:
            logger.info(
                "%s: Removing %s obsolete contacts", character, len(obsolete_ids)
            )
            self.filter(character=character, eve_entity_id__in=obsolete_ids).delete()

        create_ids = incoming_ids.difference(existing_ids)
        if create_ids:
            self._create_new_contacts(
                character=character,
                contacts_list=contacts_list,
                contact_ids=list(create_ids),
            )

        update_ids = incoming_ids.difference(create_ids)
        if update_ids:
            self._update_existing_contacts(
                character=character,
                contacts_list=contacts_list,
                contact_ids=list(update_ids),
            )

        if not obsolete_ids and not create_ids and not update_ids:
            logger.info("%s: Contacts have not changed", character)

        return create_ids

    def _create_new_contacts(self, character, contacts_list: dict, contact_ids: list):
        logger.info("%s: Storing %s new contacts", character, len(contact_ids))
        new_contacts_list = {
            contact_id: obj
            for contact_id, obj in contacts_list.items()
            if contact_id in contact_ids
        }
        new_contacts = [
            self.model(
                character=character,
                eve_entity=get_or_create_or_none("contact_id", contact_data, EveEntity),
                is_blocked=contact_data.get("is_blocked"),
                is_watched=contact_data.get("is_watched"),
                standing=contact_data.get("standing"),
            )
            for contact_data in new_contacts_list.values()
        ]
        self.bulk_create(new_contacts, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        self._update_contact_contact_labels(
            character=character,
            contacts_list=contacts_list,
            contact_ids=contact_ids,
            is_new=True,
        )

    def _update_contact_contact_labels(
        self, character, contacts_list: dict, contact_ids: list, is_new=False
    ):
        from memberaudit.models import CharacterContactLabel

        for contact_id, contact_data in contacts_list.items():
            if contact_id in contact_ids and contact_data.get("label_ids"):
                character_contact = self.filter(character=character).get(
                    eve_entity_id=contact_id
                )
                if not is_new:
                    character_contact.labels.clear()

                labels = []
                for label_id in contact_data.get("label_ids"):
                    try:
                        label = character.contact_labels.get(label_id=label_id)
                    except CharacterContactLabel.DoesNotExist:
                        # sometimes label IDs on contacts
                        # do not refer to actual labels
                        logger.info(
                            "%s: Unknown contact label with id %s",
                            character,
                            label_id,
                        )
                    else:
                        labels.append(label)

                    character_contact.labels.add(*labels)

    def _update_existing_contacts(
        self, character, contacts_list: dict, contact_ids: list
    ):
        logger.info("%s: Updating %s contacts", character, len(contact_ids))
        update_contact_pks = list(
            self.filter(character=character, eve_entity_id__in=contact_ids).values_list(
                "pk", flat=True
            )
        )
        contacts = self.in_bulk(update_contact_pks)
        for contact in contacts.values():
            contact_data = contacts_list.get(contact.eve_entity_id)
            if contact_data:
                contact.is_blocked = contact_data.get("is_blocked")
                contact.is_watched = contact_data.get("is_watched")
                contact.standing = contact_data.get("standing")

        self.bulk_update(
            contacts.values(),
            fields=["is_blocked", "is_watched", "standing"],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )
        self._update_contact_contact_labels(
            character=character, contacts_list=contacts_list, contact_ids=contact_ids
        )


class CharacterContractManager(models.Manager):
    def update_or_create_esi(self, character, force_update: bool = False):
        """Update or create contracts for a character from ESI."""
        character.update_section_if_changed(
            section=character.UpdateSection.CONTRACTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def _fetch_data_from_esi(self, character, token: Token) -> dict:
        logger.info("%s: Fetching contracts from ESI", character)
        contracts_data = esi.client.Contracts.get_characters_character_id_contracts(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        if MEMBERAUDIT_DEVELOPER_MODE:
            store_debug_data_to_disk(character, contracts_data, "contracts")

        cutoff_datetime = data_retention_cutoff()
        contracts_list = {
            obj.get("contract_id"): obj
            for obj in contracts_data
            if cutoff_datetime is None or obj.get("date_expired") > cutoff_datetime
        }
        return contracts_list

    @fetch_token_for_character("esi-universe.read_structures.v1")
    def _update_or_create_objs(
        self, character, token: Token, contracts_list
    ) -> Set[int]:
        """Update or create new contract objects from provided data.

        Return EveEntity IDs in newly created contracts.
        """
        from memberaudit.models import Location

        if cutoff_datetime := data_retention_cutoff():
            character.contracts.filter(date_expired__lt=cutoff_datetime).delete()

        existing_ids = set(character.contracts.values_list("contract_id", flat=True))
        incoming_location_ids = {
            obj["start_location_id"]
            for contract_id, obj in contracts_list.items()
            if contract_id not in existing_ids
        }
        incoming_location_ids |= {
            obj["end_location_id"] for obj in contracts_list.values()
        }
        Location.objects.create_missing_esi(incoming_location_ids, token)

        with transaction.atomic():
            incoming_ids = set(contracts_list.keys())
            existing_ids = set(
                self.filter(character=character).values_list("contract_id", flat=True)
            )
            create_ids = incoming_ids.difference(existing_ids)
            if create_ids:
                new_entity_ids = self._create_new_contracts(
                    character=character,
                    contracts_list=contracts_list,
                    contract_ids=create_ids,
                )
            else:
                new_entity_ids = set()

            update_ids = incoming_ids.difference(create_ids)
            if update_ids:
                self._update_existing_contracts(
                    character=character,
                    contracts_list=contracts_list,
                    contract_ids=update_ids,
                )
            return new_entity_ids

    def _create_new_contracts(
        self, character, contracts_list: dict, contract_ids: Set[int]
    ) -> Set[int]:
        from memberaudit.models import Location

        logger.info("%s: Storing %s new contracts", character, len(contract_ids))
        new_contracts = []
        for contract_id in contract_ids:
            contract_data = contracts_list.get(contract_id)
            if contract_data:
                new_contracts.append(
                    self.model(
                        character=character,
                        contract_id=contract_data.get("contract_id"),
                        acceptor=get_or_create_or_none(
                            "acceptor_id", contract_data, EveEntity
                        ),
                        assignee=get_or_create_or_none(
                            "assignee_id", contract_data, EveEntity
                        ),
                        availability=self.model.ESI_AVAILABILITY_MAP[
                            contract_data.get("availability")
                        ],
                        buyout=contract_data.get("buyout"),
                        collateral=contract_data.get("collateral"),
                        contract_type=self.model.ESI_TYPE_MAP.get(
                            contract_data.get("type"),
                            self.model.TYPE_UNKNOWN,
                        ),
                        date_accepted=contract_data.get("date_accepted"),
                        date_completed=contract_data.get("date_completed"),
                        date_expired=contract_data.get("date_expired"),
                        date_issued=contract_data.get("date_issued"),
                        days_to_complete=contract_data.get("days_to_complete"),
                        end_location=get_or_none(
                            "end_location_id", contract_data, Location
                        ),
                        for_corporation=contract_data.get("for_corporation"),
                        issuer_corporation=get_or_create_or_none(
                            "issuer_corporation_id", contract_data, EveEntity
                        ),
                        issuer=get_or_create_or_none(
                            "issuer_id", contract_data, EveEntity
                        ),
                        price=contract_data.get("price"),
                        reward=contract_data.get("reward"),
                        start_location=get_or_none(
                            "start_location_id", contract_data, Location
                        ),
                        status=self.model.ESI_STATUS_MAP[contract_data.get("status")],
                        title=contract_data.get("title", ""),
                        volume=contract_data.get("volume"),
                    )
                )

        self.bulk_create(new_contracts, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        return eve_entity_ids_from_objs(new_contracts)

    def _update_existing_contracts(
        self, character, contracts_list: dict, contract_ids: set
    ) -> None:
        logger.info("%s: Updating %s contracts", character, len(contract_ids))
        update_contract_pks = list(
            self.filter(character=character, contract_id__in=contract_ids).values_list(
                "pk", flat=True
            )
        )
        contracts = self.in_bulk(update_contract_pks)
        for contract in contracts.values():
            contract_data = contracts_list.get(contract.contract_id)
            if contract_data:
                contract.acceptor = get_or_create_or_none(
                    "acceptor_id", contract_data, EveEntity
                )
                contract.acceptor_corporation = get_or_create_or_none(
                    "acceptor_corporation_id", contract_data, EveEntity
                )
                contract.date_accepted = contract_data.get("date_accepted")
                contract.date_completed = contract_data.get("date_completed")
                contract.status = self.model.ESI_STATUS_MAP[contract_data.get("status")]

        self.bulk_update(
            contracts.values(),
            fields=[
                "acceptor",
                "acceptor_corporation",
                "date_accepted",
                "date_completed",
                "status",
            ],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )


class CharacterContractBidManager(models.Manager):
    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def update_or_create_esi(self, character, token: Token, contract):
        """Update or create contract bids for a contract from ESI."""
        if contract.contract_type != contract.TYPE_AUCTION:
            logger.warning(
                "%s, %s: Can not update bids. Wrong contract type.",
                character,
                contract.contract_id,
            )
            return

        logger.info(
            "%s, %s: Fetching contract bids from ESI", character, contract.contract_id
        )
        bids_data = (
            esi.client.Contracts.get_characters_character_id_contracts_contract_id_bids(
                character_id=character.eve_character.character_id,
                contract_id=contract.contract_id,
                token=token.valid_access_token(),
            ).results()
        )
        bids_list = {int(obj["bid_id"]): obj for obj in bids_data if "bid_id" in obj}
        self._update_or_create_objs(contract, bids_list)

    @transaction.atomic()
    def _update_or_create_objs(self, contract, bids_list: dict) -> Set[int]:
        """Update or create new contract objects from provided data.

        Return EveEntity IDs in newly created contracts.
        """
        incoming_ids = set(bids_list.keys())
        existing_ids = set(
            self.filter(contract=contract).values_list("bid_id", flat=True)
        )
        create_ids = incoming_ids.difference(existing_ids)
        if not create_ids:
            logger.info(
                "%s, %s: No new contract bids to add",
                contract.character,
                contract.contract_id,
            )
            return set()

        logger.info(
            "%s, %s: Storing %s new contract bids",
            contract.character,
            contract.contract_id,
            len(create_ids),
        )
        bids = [
            self.model(
                contract=contract,
                bid_id=bid.get("bid_id"),
                amount=bid.get("amount"),
                bidder=get_or_create_or_none("bidder_id", bid, EveEntity),
                date_bid=bid.get("date_bid"),
            )
            for bid_id, bid in bids_list.items()
            if bid_id in create_ids
        ]
        self.bulk_create(bids, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        new_entity_ids = {obj.bidder_id for obj in bids}
        return new_entity_ids


class CharacterContractItemQuerySet(models.QuerySet):
    def annotate_pricing(self) -> models.QuerySet:
        """Return qs with annotated price and total columns."""
        return (
            self.select_related("eve_type__market_price")
            .annotate(
                price=Case(
                    When(
                        raw_quantity=-2,
                        then=Value(None),
                    ),
                    default=F("eve_type__market_price__average_price"),
                )
            )
            .annotate(
                total=Case(
                    When(
                        raw_quantity=-2,
                        then=Value(None),
                    ),
                    default=ExpressionWrapper(
                        F("eve_type__market_price__average_price") * F("quantity"),
                        output_field=models.FloatField(),
                    ),
                )
            )
        )


class CharacterContractItemManagerBase(models.Manager):
    @fetch_token_for_character("esi-contracts.read_character_contracts.v1")
    def update_or_create_esi(self, character, token: Token, contract):
        """Update or create contract items for a contract from ESI."""
        if contract.contract_type not in [
            contract.TYPE_ITEM_EXCHANGE,
            contract.TYPE_AUCTION,
        ]:
            logger.warning(
                "%s, %s: Can not update items. Wrong contract type.",
                character,
                contract.contract_id,
            )
            return

        items_data = self._fetch_data_from_esi(character, token, contract)
        self._update_or_create_objs(contract, items_data)

    def _fetch_data_from_esi(self, character, token: Token, contract):
        logger.info(
            "%s, %s: Fetching contract items from ESI", character, contract.contract_id
        )
        my_esi = esi.client.Contracts
        items_data = my_esi.get_characters_character_id_contracts_contract_id_items(
            character_id=character.eve_character.character_id,
            contract_id=contract.contract_id,
            token=token.valid_access_token(),
        ).results()

        return items_data

    def _update_or_create_objs(self, contract, items_data):
        logger.info(
            "%s, %s: Storing %s contract items",
            self,
            contract.contract_id,
            len(items_data),
        )
        items = [
            self.model(
                contract=contract,
                record_id=item.get("record_id"),
                is_included=item.get("is_included"),
                is_singleton=item.get("is_singleton"),
                quantity=item.get("quantity"),
                raw_quantity=item.get("raw_quantity"),
                eve_type=get_or_create_esi_or_none("type_id", item, EveType),
            )
            for item in items_data
            if "record_id" in item
        ]
        with transaction.atomic():
            self.filter(contract=contract).delete()
            self.bulk_create(items, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)


CharacterContractItemManager = CharacterContractItemManagerBase.from_queryset(
    CharacterContractItemQuerySet
)



from unittest.mock import patch

from django.test import TestCase, override_settings

from app_utils.esi import EsiStatus
from app_utils.testdata_factories import UserFactory

from memberaudit.models import Character, SkillSet
from memberaudit.tests.testdata.esi_client_stub import esi_stub
from memberaudit.tests.testdata.factories import create_skill_set
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import (
    create_memberaudit_character,
    reset_celery_once_locks,
)

MANAGERS_PATH = "memberaudit.managers"
MODELS_PATH = "memberaudit.models"
TASKS_PATH = "memberaudit.tasks"


@patch(
    TASKS_PATH + ".Character.objects.get_cached",
    lambda pk, timeout: Character.objects.get(pk=pk),
)
@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_2.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_3.esi", esi_stub)
@patch(MANAGERS_PATH + ".general.esi", esi_stub)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestAdminSite(TestCase):
    fixtures = ["disable_analytics.json"]

    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        cls.user = UserFactory(is_staff=True, is_superuser=True)
        reset_celery_once_locks()

    def test_should_delete_selected_characters(self):
        # given 2 characters
        character_1001 = create_memberaudit_character(1001)
        character_1002 = create_memberaudit_character(1002)
        character_1003 = create_memberaudit_character(1003)
        self.client.force_login(self.user)

        # when selected 2 characters for deletion
        response = self.client.post(
            "/admin/memberaudit/character/",
            data={
                "action": "delete_objects",
                "select_across": 0,
                "index": 0,
                "_selected_action": [character_1001.pk, character_1002.pk],
            },
        )

        # then user is asked to confirm the 2 selected characters
        self.assertEqual(response.status_code, 200)
        text = response.content.decode("utf-8")
        self.assertIn(str(character_1001), text)
        self.assertIn(str(character_1002), text)
        self.assertNotIn(str(character_1003), text)

        # when user clicked on confirm
        response = self.client.post(
            "/admin/memberaudit/character/",
            data={
                "action": "delete_objects",
                "apply": "Delete",
                "_selected_action": [character_1001.pk, character_1002.pk],
            },
        )

        # then the selected characters are deleted, but the other character remains
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/memberaudit/character/")
        self.assertFalse(
            Character.objects.filter(
                pk__in=[character_1001.pk, character_1002.pk]
            ).exists()
        )
        self.assertTrue(Character.objects.filter(pk=character_1003.pk).exists())

    def test_should_delete_selected_skill_sets(self):
        # given 3 objects
        obj_1 = create_skill_set()
        obj_2 = create_skill_set()
        obj_3 = create_skill_set()
        self.client.force_login(self.user)

        # when user selects 2 for deletion
        response = self.client.post(
            "/admin/memberaudit/skillset/",
            data={
                "action": "delete_objects",
                "apply": "Delete",
                "_selected_action": [obj_1.pk, obj_2.pk],
            },
        )

        # then the selected objects are deleted, but the other object remains
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/admin/memberaudit/skillset/")
        self.assertFalse(SkillSet.objects.filter(pk__in=[obj_1.pk, obj_2.pk]).exists())
        self.assertTrue(SkillSet.objects.filter(pk=obj_3.pk).exists())

    def test_should_update_location_for_characters(self):
        # given 2 characters
        character_1001 = create_memberaudit_character(1001)
        self.client.force_login(self.user)

        # when user starts action
        self.client.post(
            "/admin/memberaudit/character/",
            data={
                "action": "update_section_location",
                "_selected_action": [character_1001.pk],
            },
        )

        # then character is updated
        self.assertEqual(character_1001.location.location.eve_solar_system.name, "Jita")

    def test_should_update_assets_for_characters(self):
        # given 2 characters
        character_1001 = create_memberaudit_character(1001)
        self.client.force_login(self.user)

        # when user starts action
        self.client.post(
            "/admin/memberaudit/character/",
            data={
                "action": "update_assets",
                "_selected_action": [character_1001.pk],
            },
        )

        # then character is updated
        character_1001.refresh_from_db()
        self.assertTrue(character_1001.assets.exists())



"""Character viewer views (2/2)."""
# pylint: disable=unused-argument

import datetime as dt
from collections import defaultdict
from typing import Optional

import humanize

from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Prefetch
from django.http import HttpResponse, HttpResponseNotFound, JsonResponse
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from eveuniverse.core import eveimageserver
from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.views import (
    bootstrap_icon_plus_name_html,
    bootstrap_label_html,
    link_html,
    no_wrap_html,
    yesno_str,
)

from memberaudit import __title__
from memberaudit.constants import (
    DATETIME_FORMAT,
    DEFAULT_ICON_SIZE,
    MAIL_LABEL_ID_ALL_MAILS,
    MAP_ARABIC_TO_ROMAN_NUMBERS,
    MY_DATETIME_FORMAT,
    SKILL_SET_DEFAULT_ICON_TYPE_ID,
    EveSkillTypeId,
)
from memberaudit.core.standings import Standing
from memberaudit.decorators import fetch_character_if_allowed
from memberaudit.helpers import implant_slot_num
from memberaudit.models import (
    Character,
    CharacterMail,
    CharacterPlanet,
    CharacterRole,
    CharacterStanding,
    SkillSet,
    SkillSetSkill,
)

from ._common import UNGROUPED_SKILL_SET, eve_solar_system_to_html

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

ICON_SIZE_64 = 64
CHARACTER_VIEWER_DEFAULT_TAB = "mails"

ICON_FAILED = "fas fa-times boolean-icon-false"
ICON_PARTIAL = "fas fa-check text-warning"
ICON_FULL = "fas fa-check-double text-success"
ICON_MET_ALL_REQUIRED = "fas fa-check text-success"


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_jump_clones_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Return data for character jump clones."""
    data = []
    try:
        for jump_clone in (
            character.jump_clones.select_related(
                "location",
                "location__eve_solar_system",
                "location__eve_solar_system__eve_constellation__eve_region",
            )
            .prefetch_related("implants", "implants__eve_type__dogma_attributes")
            .all()
        ):
            if (
                not jump_clone.location.is_empty
                and not jump_clone.location.is_unknown_location
            ):
                eve_solar_system = jump_clone.location.eve_solar_system
                solar_system = eve_solar_system_to_html(
                    eve_solar_system, show_region=False
                )
                region = eve_solar_system.eve_constellation.eve_region.name
            else:
                solar_system = "-"
                region = "-"

            implants_data = []
            for implant in jump_clone.implants.all():
                slot_num = implant_slot_num(implant.eve_type)
                implants_data.append(
                    {
                        "name": implant.eve_type.name,
                        "icon_url": implant.eve_type.icon_url(
                            DEFAULT_ICON_SIZE, variant=EveType.IconVariant.REGULAR
                        ),
                        "slot_num": slot_num,
                    }
                )
            if implants_data:
                implants = "<br>".join(
                    bootstrap_icon_plus_name_html(
                        implant["icon_url"], no_wrap_html(implant["name"]), size=24
                    )
                    for implant in sorted(implants_data, key=lambda k: k["slot_num"])
                )
            else:
                implants = "(none)"

            data.append(
                {
                    "id": jump_clone.pk,
                    "region": region,
                    "solar_system": solar_system,
                    "location": jump_clone.location.name_plus,
                    "implants": implants,
                }
            )
    except ObjectDoesNotExist:
        pass

    return JsonResponse({"data": data})


def _character_mail_headers_data(request, character, mail_headers_qs) -> JsonResponse:
    mails_data = []
    try:
        for mail in mail_headers_qs.select_related("sender").prefetch_related(
            "recipients"
        ):
            mail_ajax_url = reverse(
                "memberaudit:character_mail", args=[character.pk, mail.pk]
            )
            if mail.body:
                actions_html = (
                    '<button type="button" class="btn btn-primary" '
                    'data-toggle="modal" data-target="#modalCharacterMail" '
                    f"data-ajax_url={mail_ajax_url}>"
                    '<i class="fas fa-search"></i></button>'
                )
            else:
                actions_html = ""

            mails_data.append(
                {
                    "mail_id": mail.mail_id,
                    "from": mail.sender.name_plus,
                    "to": ", ".join(
                        sorted([obj.name_plus for obj in mail.recipients.all()])
                    ),
                    "subject": mail.subject,
                    "sent": mail.timestamp.isoformat(),
                    "action": actions_html,
                    "is_read": mail.is_read,
                    "is_unread_str": yesno_str(mail.is_read is False),
                }
            )
    except ObjectDoesNotExist:
        pass

    return JsonResponse({"data": mails_data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_mail_headers_by_label_data(
    request, character_pk: int, character: Character, label_id: int
) -> JsonResponse:
    """Return data for character mail headers by label."""
    if label_id == MAIL_LABEL_ID_ALL_MAILS:
        mail_headers_qs = character.mails.all()
    else:
        mail_headers_qs = character.mails.filter(labels__label_id=label_id)

    return _character_mail_headers_data(request, character, mail_headers_qs)


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_mail_headers_by_list_data(
    request, character_pk: int, character: Character, list_id: int
) -> JsonResponse:
    """Render data view for character mail headers by list."""
    mail_headers_qs = character.mails.filter(recipients__id=list_id)
    return _character_mail_headers_data(request, character, mail_headers_qs)


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_mail(
    request, character_pk: int, character: Character, mail_pk: int
) -> HttpResponse:
    """Render character mail view."""
    try:
        mail = (
            character.mails.select_related("sender")
            .prefetch_related("recipients")
            .get(pk=mail_pk)
        )
    except CharacterMail.DoesNotExist:
        error_msg = _("Mail with pk %s not found for character %s") % (
            mail_pk,
            character,
        )
        logger.warning(error_msg)
        return HttpResponseNotFound(error_msg)

    recipients = sorted(
        [
            {
                "name": obj.name_plus,
                "link": link_html(obj.external_url(), obj.name_plus),
            }
            for obj in mail.recipients.all()
        ],
        key=lambda k: k["name"],
    )
    context = {
        "mail_id": mail.mail_id,
        "labels": list(mail.labels.values_list("label_id", flat=True)),
        "sender": link_html(mail.sender.external_url(), mail.sender.name_plus),
        "recipients": format_html(", ".join([obj["link"] for obj in recipients])),
        "subject": mail.subject,
        "timestamp": mail.timestamp,
        "body": mail.body_html if mail.body else None,
        "MY_DATETIME_FORMAT": MY_DATETIME_FORMAT,
    }
    return render(
        request, "memberaudit/modals/character_viewer/mail_content.html", context
    )


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_mining_ledger_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character mining ledger."""
    qs = character.mining_ledger.select_related(
        "eve_solar_system",
        "eve_solar_system__eve_constellation__eve_region",
        "eve_type",
    ).annotate_pricing()
    data = [
        {
            "date": row.date.isoformat(),
            "quantity": row.quantity,
            "region": row.eve_solar_system.eve_constellation.eve_region.name,
            "solar_system": row.eve_solar_system.name,
            "price": row.price,
            "total": row.total,
            "type": row.eve_type.name,
        }
        for row in qs
    ]
    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_planets_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character planets."""
    data = []
    for planet in character.planets.select_related(
        "eve_planet",
        "eve_planet__eve_type",
        "eve_planet__eve_solar_system",
        "eve_planet__eve_solar_system__eve_constellation__eve_region",
    ):
        planet: CharacterPlanet
        eve_solar_system = planet.eve_planet.eve_solar_system
        solar_system_html = eve_solar_system_to_html(
            eve_solar_system, show_region=False
        )
        last_update_html = format_html(
            '<span title="{}">{}</span>',
            planet.last_update_at.strftime(DATETIME_FORMAT),
            humanize.naturaltime(planet.last_update_at),
        )
        data.append(
            {
                "id": planet.pk,
                "last_update": {
                    "display": last_update_html,
                    "sort": planet.last_update_at.isoformat(),
                },
                "num_pins": planet.num_pins,
                "region": eve_solar_system.eve_constellation.eve_region.name,
                "planet": planet.eve_planet.name,
                "solar_system": {
                    "display": solar_system_html,
                    "sort": eve_solar_system.name,
                },
                "solar_system_name": eve_solar_system.name,
                "type": planet.eve_planet.type_name(),
                "upgrade_level": MAP_ARABIC_TO_ROMAN_NUMBERS.get(
                    planet.upgrade_level, ""
                ),
            }
        )

    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_roles_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character roles."""
    character_roles_map = defaultdict(set)
    for obj in character.roles.all():
        location = CharacterRole.Location(obj.location)
        role = CharacterRole.Role(obj.role)
        character_roles_map[location].add(role)

    data = []
    if character_roles_map:
        for roles_group in CharacterRole.ROLES_GROUPED:
            location = roles_group["location"]
            group_name = roles_group["title"].title()

            for role in roles_group["roles"]:
                has_role = role in character_roles_map.get(location, [])
                data.append(
                    {
                        "group": group_name,
                        "role": role.label.title(),
                        "has_role": has_role,
                    }
                )

    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_skillqueue_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character skillqueue."""
    data = []
    try:
        for row in character.skillqueue.select_related("eve_type").filter(
            character_id=character_pk
        ):
            level_roman = MAP_ARABIC_TO_ROMAN_NUMBERS[row.finished_level]
            skill_str = f"{row.eve_type.name}&nbsp;{level_roman}"
            if row.is_active:
                skill_str += " [ACTIVE]"

            if row.finish_date:
                finish_date_humanized = humanize.naturaltime(
                    dt.datetime.now()
                    + dt.timedelta(
                        seconds=(
                            row.finish_date.timestamp() - dt.datetime.now().timestamp()
                        )
                    )
                )
                finish_date_str = (
                    f"{row.finish_date.strftime(DATETIME_FORMAT)} "
                    f"({finish_date_humanized})"
                )
                finish_date_sort = row.finish_date.isoformat()
            else:
                finish_date_str = gettext("(training not active)")
                finish_date_sort = None

            data.append(
                {
                    "position": row.queue_position + 1,
                    "skill": skill_str,
                    "finished": {
                        "display": finish_date_str,
                        "sort": finish_date_sort,
                    },
                    "is_active": row.is_active,
                }
            )
    except ObjectDoesNotExist:
        pass

    return JsonResponse({"data": data})


@login_required
@permission_required(["memberaudit.basic_access", "memberaudit.view_skill_sets"])
@fetch_character_if_allowed()
def character_skill_sets_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character skill sets."""

    def _create_row(skill_check):
        def _skill_set_name_html(skill_set):
            url = (
                skill_set.ship_type.icon_url(
                    DEFAULT_ICON_SIZE, variant=EveType.IconVariant.REGULAR
                )
                if skill_set.ship_type
                else eveimageserver.type_icon_url(
                    SKILL_SET_DEFAULT_ICON_TYPE_ID, size=DEFAULT_ICON_SIZE
                )
            )
            ship_icon = f'<img width="24" heigh="24" src="{url}"/>'
            return ship_icon + "&nbsp;&nbsp;" + skill_set.name

        def _group_name(group):
            if group:
                return (
                    group.name_plus if group.is_active else group.name + " [Not active]"
                )
            return UNGROUPED_SKILL_SET

        def _compile_failed_skills(skill_set_skills, level_name) -> Optional[list]:
            skills2 = sorted(
                [
                    {
                        "name": obj.eve_type.name,
                        "required_level": obj.required_level,
                        "recommended_level": obj.recommended_level,
                    }
                    for obj in skill_set_skills
                ],
                key=lambda k: k["name"].lower(),
            )
            return [
                bootstrap_label_html(
                    format_html(
                        "{}&nbsp;{}",
                        obj["name"],
                        MAP_ARABIC_TO_ROMAN_NUMBERS[obj[level_name]],
                    ),
                    "default",
                )
                for obj in skills2
            ]

        def _format_failed_skills(skills) -> str:
            return " ".join(skills) if skills else "-"

        failed_required_skills = list(skill_check.failed_required_skills_prefetched)
        has_required = not bool(failed_required_skills)
        failed_required_skills_str = _format_failed_skills(
            _compile_failed_skills(failed_required_skills, "required_level")
        )
        failed_recommended_skills = list(
            skill_check.failed_recommended_skills_prefetched
        )
        has_recommended = not bool(failed_recommended_skills)
        failed_recommended_skills_str = _format_failed_skills(
            _compile_failed_skills(failed_recommended_skills, "recommended_level")
        )
        is_doctrine = group.is_doctrine if group else False
        ajax_children_url = reverse(
            "memberaudit:character_skill_set_details",
            args=[character.pk, skill_check.skill_set_id],
        )
        actions_html = (
            '<button type="button" class="btn btn-primary" '
            'data-toggle="modal" data-target="#modalCharacterSkillSetDetails" '
            f"data-ajax_skill_set_detail={ ajax_children_url }>"
            '<i class="fas fa-search"></i></button>'
        )
        return {
            "id": skill_check.id,
            "group": _group_name(group),
            "skill_set": _skill_set_name_html(skill_check.skill_set),
            "skill_set_name": skill_set.name,
            "is_doctrine_str": yesno_str(is_doctrine),
            "failed_required_skills": failed_required_skills_str,
            "has_required": has_required,
            "has_required_str": yesno_str(has_required),
            "failed_recommended_skills": failed_recommended_skills_str,
            "has_recommended": has_recommended,
            "has_recommended_str": yesno_str(has_recommended),
            "action": actions_html,
        }

    groups_map = SkillSet.objects.compile_groups_map()
    skill_checks_qs = (
        character.skill_set_checks.select_related("skill_set", "skill_set__ship_type")
        .prefetch_related(
            Prefetch(
                "failed_required_skills",
                queryset=SkillSetSkill.objects.select_related("eve_type"),
                to_attr="failed_required_skills_prefetched",
            )
        )
        .prefetch_related(
            Prefetch(
                "failed_recommended_skills",
                queryset=SkillSetSkill.objects.select_related("eve_type"),
                to_attr="failed_recommended_skills_prefetched",
            )
        )
        .all()
    )
    skill_checks = {obj.skill_set_id: obj for obj in skill_checks_qs}
    data = []
    for group_map in groups_map.values():
        group = group_map["group"]
        for skill_set in group_map["skill_sets"]:
            try:
                skill_check = skill_checks[skill_set.id]
            except KeyError:
                continue
            else:
                row = _create_row(skill_check)
                data.append(row)
    data = sorted(data, key=lambda k: (k["group"].lower(), k["skill_set_name"].lower()))
    return JsonResponse({"data": data})


@login_required
@permission_required(["memberaudit.basic_access", "memberaudit.view_skill_sets"])
@fetch_character_if_allowed()
def character_skill_set_details(
    request, character_pk: int, character: Character, skill_set_pk: int
) -> HttpResponse:
    """Render view for character skill set details."""

    def _compile_row(character_skills, missing_skills, skill_id, skill):
        character_skill = character_skills.get(skill_id)
        recommended_level_str = "-"
        required_level_str = "-"
        current_str = "-"
        result_icon = ICON_FAILED
        met_required = True

        if character_skill:
            current_str = MAP_ARABIC_TO_ROMAN_NUMBERS[
                character_skill.active_skill_level
            ]

        if skill.recommended_level:
            recommended_level_str = MAP_ARABIC_TO_ROMAN_NUMBERS[skill.recommended_level]

        if skill.required_level:
            required_level_str = MAP_ARABIC_TO_ROMAN_NUMBERS[skill.required_level]

        if not character_skill:
            result_icon = ICON_FAILED
            met_required = False
        else:
            if (
                skill.required_level
                and not skill.recommended_level
                and character_skill.active_skill_level >= skill.required_level
            ):
                result_icon = ICON_FULL
            elif (
                skill.recommended_level
                and character_skill.active_skill_level >= skill.recommended_level
            ):
                result_icon = ICON_FULL
            elif (
                skill.required_level
                and character_skill.active_skill_level >= skill.required_level
            ):
                result_icon = ICON_PARTIAL
            else:
                met_required = False

        if not character_skill or (
            character_skill and character_skill.active_skill_level < skill.maximum_level
        ):
            missing_skills.append(skill.maximum_skill_str)

        return {
            "name": skill.eve_type.name,
            "required": required_level_str,
            "recommended": recommended_level_str,
            "current": current_str,
            "result_icon": result_icon,
            "met_required": met_required,
        }

    skill_set = get_object_or_404(SkillSet, pk=skill_set_pk)
    skill_set_skills = _calc_skill_set_skills(skill_set_pk)
    character_skills = _calc_character_skills(character, skill_set_skills)
    out_data = []
    missing_skills = []
    for skill_id, skill in skill_set_skills.items():
        out_data.append(_compile_row(character_skills, missing_skills, skill_id, skill))

    met_all_required = True
    for data in out_data:
        if not data["met_required"]:
            met_all_required = False
            break

    context = {
        "name": skill_set.name,
        "description": skill_set.description,
        "ship_url": _calc_url_for_skill_set(skill_set),
        "skills": sorted(out_data, key=lambda k: (k["name"].lower())),
        "met_all_required": met_all_required,
        "icon_failed": ICON_FAILED,
        "icon_partial": ICON_PARTIAL,
        "icon_full": ICON_FULL,
        "icon_met_all_required": ICON_MET_ALL_REQUIRED,
        "missing_skills_str": "\n".join(missing_skills) if missing_skills else "",
    }

    return render(
        request,
        "memberaudit/modals/character_viewer/skill_set_content.html",
        context,
    )


def _calc_url_for_skill_set(skill_set):
    url = (
        skill_set.ship_type.icon_url(ICON_SIZE_64, variant=EveType.IconVariant.REGULAR)
        if skill_set.ship_type
        else eveimageserver.type_icon_url(
            SKILL_SET_DEFAULT_ICON_TYPE_ID, size=ICON_SIZE_64
        )
    )

    return url


def _calc_skill_set_skills(skill_set_pk):
    skill_set_skills_qs = SkillSetSkill.objects.select_related("eve_type").filter(
        skill_set_id=skill_set_pk
    )
    skill_set_skills = {obj.eve_type_id: obj for obj in skill_set_skills_qs}
    return skill_set_skills


def _calc_character_skills(character, skill_set_skills):
    character_skills_qs = character.skills.select_related("eve_type").filter(
        eve_type_id__in=skill_set_skills.keys()
    )
    character_skills = {obj.eve_type_id: obj for obj in character_skills_qs}
    return character_skills


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_skills_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character skills."""
    skills_data = []
    try:
        for skill in character.skills.select_related("eve_type", "eve_type__eve_group"):
            level_str = MAP_ARABIC_TO_ROMAN_NUMBERS[skill.active_skill_level]
            skill_name = format_html(
                '<span title="{}">{} {}</span>',
                skill.eve_type.description,
                skill.eve_type.name,
                level_str,
            )
            skills_data.append(
                {
                    "group": skill.eve_type.eve_group.name,
                    "skill": skill.eve_type.name,
                    "skill_name": skill_name,
                    "level": skill.active_skill_level,
                    "level_str": level_str,
                }
            )
    except ObjectDoesNotExist:
        pass

    return JsonResponse({"data": skills_data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_standings_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character standings."""
    connections_skill_level = character.skills.find_active_skill_level(
        EveSkillTypeId.CONNECTIONS
    )
    diplomacy_skill_level = character.skills.find_active_skill_level(
        EveSkillTypeId.DIPLOMACY
    )
    criminal_connections_skill_level = character.skills.find_active_skill_level(
        EveSkillTypeId.CRIMINAL_CONNECTIONS
    )
    data = []
    for obj in character.standings.select_related("eve_entity").all():
        obj: CharacterStanding
        name = obj.eve_entity.name
        name_html = bootstrap_icon_plus_name_html(
            obj.eve_entity.icon_url(DEFAULT_ICON_SIZE), name, avatar=True
        )
        map_category = {
            "character": _("Agent"),
            "corporation": _("Corporation"),
            "faction": _("Faction"),
        }
        npc_type = map_category.get(obj.eve_entity.get_category_display(), "")
        effective_standing = obj.effective_standing(
            connections_skill_level=connections_skill_level,
            criminal_connections_skill_level=criminal_connections_skill_level,
            diplomacy_skill_level=diplomacy_skill_level,
        )
        standing_str = f"{effective_standing:.2f} ({obj.standing:.2f})"
        standing_group = Standing.from_value(effective_standing)
        data.append(
            {
                "id": obj.eve_entity.id,
                "group_name": standing_group.label.title(),
                "group_sort": standing_group.value,
                "name": {"display": name_html, "sort": name},
                "standing": {"display": standing_str, "sort": effective_standing},
                "type": npc_type,
            }
        )

    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_titles_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character titles."""
    data = [
        {"id": title.title_id, "name": title.name}
        for title in character.titles.order_by("title_id")
    ]
    return JsonResponse({"data": data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_wallet_journal_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character wallet journal."""
    wallet_data = []
    try:
        for row in character.wallet_journal.select_related(
            "first_party", "second_party"
        ).all():
            first_party = row.first_party.name if row.first_party else "-"
            second_party = row.second_party.name if row.second_party else "-"
            wallet_data.append(
                {
                    "date": row.date.isoformat(),
                    "ref_type": row.ref_type.replace("_", " ").title(),
                    "first_party": first_party,
                    "second_party": second_party,
                    "amount": float(row.amount),
                    "balance": float(row.balance),
                    "description": row.description,
                    "reason": row.reason,
                }
            )
    except ObjectDoesNotExist:
        pass

    return JsonResponse({"data": wallet_data})


@login_required
@permission_required("memberaudit.basic_access")
@fetch_character_if_allowed()
def character_wallet_transactions_data(
    request, character_pk: int, character: Character
) -> JsonResponse:
    """Render data view for character wallet transactions."""
    wallet_data = []
    try:
        for row in character.wallet_transactions.select_related(
            "client", "eve_type", "location"
        ).all():
            buy_or_sell = _("Buy") if row.is_buy else _("Sell")
            wallet_data.append(
                {
                    "date": row.date.isoformat(),
                    "quantity": row.quantity,
                    "type": row.eve_type.name,
                    "unit_price": float(row.unit_price),
                    "total": float(
                        row.unit_price * row.quantity * (-1 if row.is_buy else 1)
                    ),
                    "client": row.client.name,
                    "location": row.location.name,
                    "is_buy": row.is_buy,
                    "buy_or_sell": buy_or_sell,
                }
            )
    except ObjectDoesNotExist:
        pass
    return JsonResponse({"data": wallet_data})



"""New style asset tests.

These tests do not use the esi testdata fixtures.
"""

from unittest.mock import patch

from django.test import TestCase, override_settings
from eveuniverse.models import EveSolarSystem

from app_utils.esi_testing import EsiClientStub, EsiEndpoint

from memberaudit import tasks
from memberaudit.models import Character, CharacterAsset, Location
from memberaudit.tests.testdata.constants import EveTypeId
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    create_memberaudit_character,
    reset_celery_once_locks,
)

from .testdata.factories import (
    create_character_asset,
    create_character_location,
    create_character_ship,
    create_location_eve_solar_system,
)

MODELS_PATH = "memberaudit.models"
MANAGERS_PATH = "memberaudit.managers"
TASKS_PATH = "memberaudit.tasks"


@patch(MANAGERS_PATH + ".character_sections_1.esi")
class TestUpdateCharacterAssetsBuildListFromEsi(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.jita = EveSolarSystem.objects.get(name="Jita")
        cls.location_jita_44 = Location.objects.get(id=60003760)
        cls.amamake = EveSolarSystem.objects.get(name="Amamake")
        cls.location_structure_1 = Location.objects.get(id=1_000_000_000_001)
        cls.location_jita = create_location_eve_solar_system(id=cls.jita.id)
        cls.item_ids = {1_100_000_000_001}
        cls.endpoints = [
            EsiEndpoint(
                "Assets",
                "get_characters_character_id_assets",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": True,
                            "item_id": 1_100_000_000_001,
                            "location_flag": "Hangar",
                            "location_id": cls.location_jita_44.id,
                            "location_type": "station",
                            "quantity": 1,
                            "type_id": EveTypeId.VELDSPAR,
                        }
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_characters_character_id_assets_names",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {"item_id": 1_100_000_000_001, "name": "ESI asset"},
                    ]
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    def test_should_add_current_ship_when_it_not_in_assets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        create_character_ship(
            character=self.character_1001,
            item_id=1_100_000_000_999,
            eve_type_id=EveTypeId.MERLIN,
            name="Joy Ride",
        )
        create_character_location(
            character=self.character_1001, location=self.location_jita_44
        )

        # when
        result = tasks.assets_build_list_from_esi(self.character_1001.pk)

        # then
        asset_data = {asset["item_id"]: asset for asset in result}
        self.assertIn(1_100_000_000_999, asset_data.keys())

    def test_should_not_add_current_ship_when_not_generated(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub

        # when
        result = tasks.assets_build_list_from_esi(self.character_1001.pk)

        # then
        item_ids = {asset["item_id"] for asset in result}
        self.assertSetEqual(item_ids, {1_100_000_000_001})

    def test_should_not_add_current_ship_when_already_in_assets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        create_character_ship(
            character=self.character_1001,
            item_id=1_100_000_000_001,
            eve_type_id=EveTypeId.MERLIN,
            name="Joy Ride",
        )
        create_character_location(
            character=self.character_1001, location=self.location_jita
        )

        # when
        result = tasks.assets_build_list_from_esi(self.character_1001.pk)

        # then
        asset_data = {asset["item_id"]: asset for asset in result}
        obj = asset_data[1_100_000_000_001]
        self.assertNotEqual(obj["name"], "Joy Ride")

    def test_should_return_none_when_asset_list_is_unchanged_wo_ship(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        tasks.assets_build_list_from_esi(self.character_1001.pk)

        # when
        result = tasks.assets_build_list_from_esi(self.character_1001.pk)

        # then
        self.assertIsNone(result)

    def test_should_return_none_when_asset_list_is_unchanged_w_ship(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        create_character_ship(
            character=self.character_1001,
            item_id=1_100_000_000_999,
            eve_type_id=EveTypeId.MERLIN,
            name="Joy Ride",
        )
        create_character_location(
            character=self.character_1001, location=self.location_jita
        )
        tasks.assets_build_list_from_esi(self.character_1001.pk)

        # when
        result = tasks.assets_build_list_from_esi(self.character_1001.pk)

        # then
        self.assertIsNone(result)


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(MANAGERS_PATH + ".character_sections_1.esi")
class TestUpdateCharacterAssets2(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        reset_celery_once_locks()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.jita = EveSolarSystem.objects.get(id=30000142)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.amamake = EveSolarSystem.objects.get(id=30002537)
        cls.structure_1 = Location.objects.get(id=1_000_000_000_001)
        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_characters_character_id_assets",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": True,
                            "item_id": 1_100_000_000_001,
                            "location_flag": "Hangar",
                            "location_id": cls.jita_44.id,
                            "location_type": "station",
                            "quantity": 1,
                            "type_id": EveTypeId.CHARON.value,
                        },
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": False,
                            "item_id": 1_100_000_000_002,
                            "location_flag": "Hangar",
                            "location_id": cls.jita_44.id,
                            "location_type": "station",
                            "quantity": 1,
                            "type_id": EveTypeId.VELDSPAR.value,
                        },
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": False,
                            "item_id": 1_100_000_000_003,
                            "location_flag": "Hangar",
                            "location_id": 1_100_000_000_001,  # Charon
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": EveTypeId.CARGO_CONTAINER.value,
                        },
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": False,
                            "item_id": 1_100_000_000_004,
                            "location_flag": "???",
                            "location_id": 1_100_000_000_003,  # Cargo container
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": EveTypeId.VELDSPAR.value,
                        },
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": True,
                            "item_id": 1_100_000_000_005,
                            "location_flag": "???",
                            "location_id": 1_100_000_000_003,  # Cargo container
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": EveTypeId.MERLIN.value,
                        },
                        {
                            "is_blueprint_copy": False,
                            "is_singleton": False,
                            "item_id": 1_100_000_000_006,
                            "location_flag": "???",
                            "location_id": 1_100_000_000_005,  # Merlin
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": EveTypeId.VELDSPAR.value,
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_characters_character_id_assets_names",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {"item_id": 1_100_000_000_001, "name": "Freighter"},
                        {"item_id": 1_100_000_000_005, "name": "Fighter"},
                    ]
                },
            ),
        ]
        cls.esi_client = EsiClientStub.create_from_endpoints(endpoints)

    def test_should_create_assets_from_scratch(self, mock_esi):
        # given
        mock_esi.client = self.esi_client

        # when
        tasks.update_character_assets.delay(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
            },
        )

        obj: CharacterAsset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        self.assertFalse(obj.is_blueprint_copy)
        self.assertTrue(obj.is_singleton)
        self.assertEqual(obj.location_flag, "Hangar")
        self.assertEqual(obj.location.id, self.jita_44.id)
        self.assertEqual(obj.quantity, 1)
        self.assertEqual(obj.eve_type.id, EveTypeId.CHARON)
        self.assertEqual(obj.name, "Freighter")

    def test_should_remove_obsolete_assets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client
        create_character_asset(
            character=self.character_1001, item_id=1100000000666, location=self.jita_44
        )

        # when
        tasks.update_character_assets.delay(self.character_1001.pk, True)

        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
            },
        )
        status = self.character_1001.update_status_for_section(
            Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)

    def test_should_update_existing_assets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client
        create_character_asset(
            character=self.character_1001,
            item_id=1_100_000_000_002,
            location=self.jita_44,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            is_singleton=False,
            quantity=10,
        )

        # when
        tasks.update_character_assets.delay(self.character_1001.pk, True)

        # then
        obj: CharacterAsset = self.character_1001.assets.get(item_id=1_100_000_000_002)
        self.assertEqual(obj.quantity, 1)
        status = self.character_1001.update_status_for_section(
            Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)

    @patch(TASKS_PATH + ".logger", wraps=tasks.logger)
    def test_log_warning_when_there_are_leftovers_1(self, mock_logger, mock_esi):
        # given
        asset_data = {
            1_100_000_000_001: {
                "is_blueprint_copy": False,
                "is_singleton": False,
                "item_id": 1_100_000_000_001,
                "location_flag": "Hangar",
                "location_id": self.jita_44.id,
                "location_type": "station",
                "quantity": 1,
                "type_id": EveTypeId.VELDSPAR.value,
            },
            1_100_000_000_002: {
                "is_blueprint_copy": False,
                "is_singleton": True,
                "item_id": 1_100_000_000_002,
                "location_flag": "Hangar",
                "location_id": self.jita_44.id,
                "location_type": "station",
                "quantity": 1,
                "type_id": EveTypeId.CHARON.value,
            },
            1_100_000_000_003: {
                "is_blueprint_copy": False,
                "is_singleton": False,
                "item_id": 1_100_000_000_003,
                "location_flag": "Hangar",
                "location_id": 1_100_000_000_009,  # Unknown location
                "location_type": "item",
                "quantity": 1,
                "type_id": EveTypeId.VELDSPAR.value,
            },
        }

        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_characters_character_id_assets",
                "character_id",
                needs_token=True,
                data={"1001": list(asset_data.values())},
            ),
            EsiEndpoint(
                "Assets",
                "post_characters_character_id_assets_names",
                "character_id",
                needs_token=True,
                data={
                    "1001": [
                        {"item_id": 1_100_000_000_002, "name": "Freighter"},
                    ]
                },
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        with patch(TASKS_PATH + ".Character.assets_preload_objects", spec=True) as _:
            tasks.update_character_assets.delay(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {1_100_000_000_001, 1_100_000_000_002},
        )
        self.assertTrue(mock_logger.warning.called)
        status = self.character_1001.update_status_for_section(
            Character.UpdateSection.ASSETS
        )
        self.assertFalse(status.is_success)

    @patch(TASKS_PATH + ".MEMBERAUDIT_TASKS_MAX_ASSETS_PER_PASS", 1)
    @patch(TASKS_PATH + ".assets_create_children", wraps=tasks.assets_create_children)
    @patch(
        TASKS_PATH + "._assets_create_parents_chunk",
        wraps=tasks._assets_create_parents_chunk,
    )
    def test_should_create_parent_and_child_assets_in_chunks_when_too_many(
        self, mock_assets_create_parents_chunk, mock_assets_create_children, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client

        # when
        tasks.update_character_assets.delay(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
            },
        )
        self.assertEqual(len(mock_assets_create_parents_chunk.mock_calls), 2)
        self.assertEqual(len(mock_assets_create_children.mock_calls), 4)

    def test_should_create_parent_assets_only(self, mock_esi):
        # given
        asset_data = {
            1_100_000_000_001: {
                "is_blueprint_copy": False,
                "is_singleton": False,
                "item_id": 1_100_000_000_001,
                "location_flag": "Hangar",
                "location_id": self.jita_44.id,
                "location_type": "station",
                "quantity": 1,
                "type_id": EveTypeId.VELDSPAR.value,
            }
        }

        endpoints = [
            EsiEndpoint(
                "Assets",
                "get_characters_character_id_assets",
                "character_id",
                needs_token=True,
                data={"1001": list(asset_data.values())},
            ),
            EsiEndpoint(
                "Assets",
                "post_characters_character_id_assets_names",
                "character_id",
                needs_token=True,
                data={"1001": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        with patch(TASKS_PATH + ".Character.assets_preload_objects", spec=True) as _:
            tasks.update_character_assets.delay(self.character_1001.pk, True)

        # then
        self.assertSetEqual(self.character_1001.assets.item_ids(), {1_100_000_000_001})
        status = self.character_1001.update_status_for_section(
            Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)



"""Helpers for Member Audit."""

import datetime as dt
import itertools
import json
import os
from typing import Any, Iterable, Optional, Set

from celery import Task

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveType

from app_utils.datetime import datetime_round_hour

from memberaudit.app_settings import MEMBERAUDIT_DATA_RETENTION_LIMIT
from memberaudit.constants import EveDogmaAttributeId


class EveEntityIdsMixin:
    """Add EveEntity related features."""

    def eve_entity_ids(self) -> Set[int]:
        """Return IDs of all existing EveEntity objects for this object."""
        ids = set()
        for field in self._meta.get_fields():
            if field.is_relation and issubclass(field.related_model, EveEntity):
                value = getattr(self, field.attname)
                if value:
                    ids.add(value)
        return ids


def eve_entity_ids_from_objs(objs: Iterable[Any]) -> Set[int]:
    """Return all EveEntity IDs from objs. Will return an empty set when objs is empty.

    Expects objs to have the `EveEntityIdsMixin`.
    """
    if not objs:
        return set()

    entity_ids_list = [obj.eve_entity_ids() for obj in objs]
    return set(itertools.chain.from_iterable(entity_ids_list))


def data_retention_cutoff() -> Optional[dt.datetime]:
    """Return cutoff datetime for data retention of None if unlimited."""
    if MEMBERAUDIT_DATA_RETENTION_LIMIT is None:
        return None
    return datetime_round_hour(
        now() - dt.timedelta(days=MEMBERAUDIT_DATA_RETENTION_LIMIT)
    )


def store_debug_data_to_disk(character, lst: Any, name: str):
    """Store character related data as JSON file to disk (for debugging).

    Will store under memberaudit_logs/{DATE}/{CHARACTER_PK}_{NAME}.json
    """
    today_str = now().strftime("%Y%m%d")
    now_str = now().strftime("%Y%m%d%H%M")
    path = f"memberaudit_log/{today_str}"
    if not os.path.isdir(path):
        os.makedirs(path)

    fullpath = os.path.join(path, f"character_{character.pk}_{name}_{now_str}.json")
    try:
        with open(fullpath, "w", encoding="utf-8") as file:
            json.dump(lst, file, cls=DjangoJSONEncoder, sort_keys=True, indent=4)

    except OSError:
        pass


def implant_slot_num(implant_type: EveType) -> int:  # TODO: Refactor into model
    """Return slot number for an implant. Or 0 if not found."""
    dogma_attributes = {
        obj.eve_dogma_attribute_id: obj.value
        for obj in implant_type.dogma_attributes.all()
    }
    try:
        slot_num = int(dogma_attributes[EveDogmaAttributeId.IMPLANT_SLOT])
    except KeyError:
        slot_num = 0
    return slot_num


def determine_task_priority(task_obj: Task) -> Optional[int]:
    """Return priority of give task or None if not defined."""
    properties = task_obj.request.get("properties") or {}
    return properties.get("priority")



from unittest.mock import patch

from django.utils.dateparse import parse_datetime
from eveuniverse.models import EveEntity

from app_utils.testing import NoSocketsTestCase

from memberaudit.models import MailEntity
from memberaudit.tests.testdata.esi_client_stub import esi_client_stub
from memberaudit.tests.testdata.factories import (
    create_character_mail,
    create_mailing_list,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import create_memberaudit_character

MANAGERS_PATH = "memberaudit.managers"
MODELS_PATH = "memberaudit.models"
TASKS_PATH = "memberaudit.tasks"


@patch(MANAGERS_PATH + ".character_sections_2.esi")
@patch(MANAGERS_PATH + ".general.esi")
class TestCharacterMailUpdate(NoSocketsTestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        load_entities()
        cls.character = create_memberaudit_character(1001)

    @staticmethod
    def stub_eve_entity_get_or_create_esi(id, *args, **kwargs):
        """will return EveEntity if it exists else None, False"""
        try:
            obj = EveEntity.objects.get(id=id)
            return obj, True
        except EveEntity.DoesNotExist:
            return None, False

    @patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
    @patch(MANAGERS_PATH + ".character_sections_2.EveEntity.objects.get_or_create_esi")
    def test_should_update_existing_mail_headers(
        self,
        mock_eve_entity,
        mock_esi_character,
        mock_esi_sections,
    ):
        # given
        mock_esi_character.client = esi_client_stub
        mock_esi_sections.client = esi_client_stub
        mock_eve_entity.side_effect = self.stub_eve_entity_get_or_create_esi
        sender, _ = MailEntity.objects.update_or_create_from_eve_entity_id(id=1002)
        mail = create_character_mail(
            character=self.character,
            mail_id=1,
            sender=sender,
            subject="Mail 1",
            body="My body text",
            timestamp=parse_datetime("2015-09-05T16:07:00Z"),
            is_read=False,  # to be updated
        )
        recipient_1, _ = MailEntity.objects.update_or_create_from_eve_entity_id(id=1001)
        recipient_2 = create_mailing_list()
        mail.recipients.set([recipient_1, recipient_2])

        self.character.update_mailing_lists()
        self.character.update_mail_labels()

        label = self.character.mail_labels.get(label_id=17)
        mail.labels.add(label)  # to be updated

        # when
        self.character.update_mail_headers()

        # then
        self.assertSetEqual(
            set(self.character.mails.values_list("mail_id", flat=True)),
            {1, 2, 3},
        )

        obj = self.character.mails.get(mail_id=1)
        self.assertEqual(obj.sender_id, 1002)
        self.assertTrue(obj.is_read)
        self.assertEqual(obj.subject, "Mail 1")
        self.assertEqual(obj.timestamp, parse_datetime("2015-09-05T16:07:00Z"))
        self.assertEqual(obj.body, "My body text")
        recipient_ids = set(obj.recipients.values_list("id", flat=True))
        self.assertSetEqual(recipient_ids, {recipient_1.id, recipient_2.id})
        label_ids = set(obj.labels.values_list("label_id", flat=True))
        self.assertSetEqual(label_ids, {3})
